from __future__ import annotations

import ctypes

try:
    import fcntl
except ImportError:
    # fcntl is not available on Windows
    fcntl = None  # type: ignore
import logging
import socket
from typing import Any

_LOGGER = logging.getLogger(__name__)

AF_BLUETOOTH = 31
PF_BLUETOOTH = AF_BLUETOOTH
BTPROTO_HCI = 1
HCI_MAX_DEV = 16
HCIGETDEVLIST = 0x800448D2  # _IOR('H', 210, int)
HCIGETDEVINFO = 0x800448D3  # _IOR('H', 211, int)

HCI_UP = 0  # HCI device flag
LMP_LE = 0x40  # LMP capabilities


class hci_dev_req(ctypes.Structure):
    _fields_ = [("dev_id", ctypes.c_uint16), ("dev_opt", ctypes.c_uint32)]


class hci_dev_list_req(ctypes.Structure):
    _fields_ = [("dev_num", ctypes.c_uint16), ("dev_req", hci_dev_req * HCI_MAX_DEV)]


class bdaddr_t(ctypes.Structure):
    _fields_ = [("b", ctypes.c_uint8 * 6)]

    def __str__(self) -> str:
        return ":".join(["%02X" % x for x in reversed(self.b)])


class hci_dev_stats(ctypes.Structure):
    _fields_ = [
        ("err_rx", ctypes.c_uint32),
        ("err_tx", ctypes.c_uint32),
        ("cmd_tx", ctypes.c_uint32),
        ("evt_rx", ctypes.c_uint32),
        ("acl_tx", ctypes.c_uint32),
        ("acl_rx", ctypes.c_uint32),
        ("sco_tx", ctypes.c_uint32),
        ("sco_rx", ctypes.c_uint32),
        ("byte_rx", ctypes.c_uint32),
        ("byte_tx", ctypes.c_uint32),
    ]


class hci_dev_info(ctypes.Structure):
    _fields_ = [
        ("dev_id", ctypes.c_uint16),
        ("name", ctypes.c_char * 8),
        ("bdaddr", bdaddr_t),
        ("flags", ctypes.c_uint32),
        ("type", ctypes.c_uint8),
        ("features", ctypes.c_uint8 * 8),
        ("pkt_type", ctypes.c_uint32),
        ("link_policy", ctypes.c_uint32),
        ("link_mode", ctypes.c_uint32),
        ("acl_mtu", ctypes.c_uint16),
        ("acl_pkts", ctypes.c_uint16),
        ("sco_mtu", ctypes.c_uint16),
        ("sco_pkts", ctypes.c_uint16),
        ("stat", hci_dev_stats),
    ]


hci_dev_info_p = ctypes.POINTER(hci_dev_info)


def get_adapters_from_hci() -> dict[int, dict[str, Any]]:
    """Get bluetooth adapters from HCI."""
    if not fcntl:
        raise RuntimeError("fcntl is not available")
    out: dict[int, dict[str, Any]] = {}
    sock: socket.socket | None = None
    try:
        sock = socket.socket(AF_BLUETOOTH, socket.SOCK_RAW, BTPROTO_HCI)
        buf = hci_dev_list_req()
        buf.dev_num = HCI_MAX_DEV
        ret = fcntl.ioctl(sock.fileno(), HCIGETDEVLIST, buf)
        if ret < 0:
            raise OSError(f"HCIGETDEVLIST failed: {ret}")
        for i in range(buf.dev_num):
            dev_req = buf.dev_req[i]
            dev = hci_dev_info()
            dev.dev_id = dev_req.dev_id
            ret = fcntl.ioctl(sock.fileno(), HCIGETDEVINFO, dev)
            info = {str(k): getattr(dev, k) for k, *v_ in dev._fields_}
            info["bdaddr"] = str(info["bdaddr"])
            info["name"] = info["name"].decode()
            info["powered"] = bool(info["flags"] & (1 << HCI_UP))
            info["advertise"] = bool(
                info["features"][4] & LMP_LE
            )  # assume advertising supported if device supports LE
            out[int(dev.dev_id)] = info
    except OSError as error:
        _LOGGER.debug("Error while getting HCI devices: %s", error)
        return out
    except Exception as error:  # pylint: disable=broad-except
        _LOGGER.exception("Unexpected error while getting HCI devices: %s", error)
        return out
    finally:
        if sock:
            sock.close()
    return out



from __future__ import annotations

import asyncio
import logging
import platform
from typing import Any

from uart_devices import BluetoothDevice as UARTBluetoothDevice
from uart_devices import NotAUARTDeviceError
from usb_devices import BluetoothDevice as USBBluetoothDevice
from usb_devices import NotAUSBDeviceError

import aiooui

from ..adapters import BluetoothAdapters
from ..const import EMPTY_MAC_ADDRESS, UNIX_DEFAULT_BLUETOOTH_ADAPTER
from ..dbus import BlueZDBusObjects
from ..history import AdvertisementHistory
from ..models import AdapterDetails
from .linux_hci import get_adapters_from_hci

_LOGGER = logging.getLogger(__name__)


class LinuxAdapters(BluetoothAdapters):
    """Class for getting the bluetooth adapters on a Linux system."""

    def __init__(self) -> None:
        """Initialize the adapter."""
        self._bluez = BlueZDBusObjects()
        self._adapters: dict[str, AdapterDetails] | None = None
        self._devices: dict[str, UARTBluetoothDevice | USBBluetoothDevice] = {}
        self._hci_output: dict[int, dict[str, Any]] | None = None

    async def refresh(self) -> None:
        """Refresh the adapters."""
        loop = asyncio.get_running_loop()
        load_task = asyncio.create_task(self._bluez.load())
        adapters_from_hci_future = loop.run_in_executor(None, get_adapters_from_hci)
        futures: list[asyncio.Future[Any]] = [load_task, adapters_from_hci_future]
        if not aiooui.is_loaded():
            futures.append(aiooui.async_load())
        await asyncio.gather(*futures)
        self._hci_output = await adapters_from_hci_future
        self._adapters = None  # clear cache
        self._devices = {}
        await loop.run_in_executor(None, self._refresh_devices)

    def _refresh_devices(self) -> None:
        """Refresh the devices."""
        for adapter in self._bluez.adapter_details:
            i = int(adapter[3:])
            for cls in (USBBluetoothDevice, UARTBluetoothDevice):
                dev = cls(i)
                self._devices[adapter] = dev
                try:
                    dev.setup()
                except (NotAUARTDeviceError, NotAUSBDeviceError):
                    continue
                except FileNotFoundError:
                    break
                except Exception:  # pylint: disable=broad-except
                    _LOGGER.exception("Unexpected error setting up device hci%s", dev)
                    break
                else:
                    break

    @property
    def history(self) -> dict[str, AdvertisementHistory]:
        """Get the bluez history."""
        return self._bluez.history

    @property
    def adapters(self) -> dict[str, AdapterDetails]:
        """Get the adapter details."""
        manufacturer: str | None
        if self._adapters is None:
            adapters: dict[str, AdapterDetails] = {}
            if self._hci_output:
                for hci_details in self._hci_output.values():
                    name = hci_details["name"]
                    mac_address = hci_details["bdaddr"].upper()
                    if mac_address == EMPTY_MAC_ADDRESS:
                        manufacturer = None
                    else:
                        manufacturer = aiooui.get_vendor(mac_address)
                    adapters[name] = AdapterDetails(
                        address=mac_address,
                        sw_version=platform.release(),
                        hw_version=None,
                        passive_scan=False,  # assume false if we don't know
                        manufacturer=manufacturer,
                        product=None,
                        vendor_id=None,
                        product_id=None,
                        adapter_type=None,  # Unknown for hci-only adapters
                        advertise=hci_details["advertise"],
                        powered=hci_details["powered"],
                    )
            adapter_details = self._bluez.adapter_details
            for adapter, details in adapter_details.items():
                if not (adapter1 := details.get("org.bluez.Adapter1")):
                    continue
                mac_address = adapter1["Address"]
                device = self._devices[adapter]
                product: str | None = None
                manufacturer = None
                vendor_id: str | None = None
                product_id: str | None = None
                adapter_type: str | None = None
                powered = adapter1["Powered"]
                if isinstance(device, USBBluetoothDevice):
                    adapter_type = "usb"
                    usb_device = device.usb_device
                    if mac_address != EMPTY_MAC_ADDRESS and (
                        usb_device is None
                        or usb_device.vendor_id == usb_device.manufacturer
                        or usb_device.manufacturer is None
                        or usb_device.manufacturer == "Unknown"
                    ):
                        manufacturer = aiooui.get_vendor(mac_address)
                    elif usb_device is not None:
                        manufacturer = usb_device.manufacturer
                    if usb_device is not None:
                        product = usb_device.product
                        vendor_id = usb_device.vendor_id
                        product_id = usb_device.product_id
                elif isinstance(device, UARTBluetoothDevice):
                    adapter_type = "uart"
                    uart_device = device.uart_device
                    if uart_device is None:
                        if mac_address != EMPTY_MAC_ADDRESS:
                            manufacturer = aiooui.get_vendor(mac_address)
                    else:
                        product = uart_device.product

                        if mac_address == EMPTY_MAC_ADDRESS:
                            manufacturer = uart_device.manufacturer
                        else:
                            manufacturer = (
                                aiooui.get_vendor(mac_address)
                                or uart_device.manufacturer
                            )

                adapters[adapter] = AdapterDetails(
                    address=mac_address,
                    sw_version=platform.release(),
                    hw_version=adapter1.get("Modalias"),
                    passive_scan="org.bluez.AdvertisementMonitorManager1" in details,
                    manufacturer=manufacturer,
                    product=product,
                    vendor_id=vendor_id,
                    product_id=product_id,
                    adapter_type=adapter_type,
                    advertise="org.bluez.LEAdvertisingManager1" in details,
                    powered=powered,
                )
            self._adapters = adapters
        return self._adapters

    @property
    def default_adapter(self) -> str:
        """Get the default adapter."""
        return UNIX_DEFAULT_BLUETOOTH_ADAPTER



import asyncio
import time
from platform import system
from typing import Any
from unittest.mock import ANY, MagicMock, patch

import pytest
from bleak.backends.device import BLEDevice
from bleak.backends.scanner import AdvertisementData

try:
    from dbus_fast import AuthError, MessageType
except (AttributeError, ImportError):
    MessageType = None
    AuthError = None
    # dbus_fast is not available on Windows
import bluetooth_adapters.dbus as bluetooth_adapters_dbus
from bluetooth_adapters import get_manufacturer_from_mac
from uart_devices import BluetoothDevice as UARTBluetoothDevice
from uart_devices import NotAUARTDeviceError, UARTDevice
from usb_devices import BluetoothDevice as USBBluetoothDevice
from usb_devices import NotAUSBDeviceError, USBDevice

if system() != "Windows":
    from bluetooth_adapters import (
        BlueZDBusObjects,
        get_bluetooth_adapters,
        get_dbus_managed_objects,
    )
else:
    BlueZDBusObjects = None  # type: ignore
    get_bluetooth_adapters = None  # type: ignore
    get_dbus_managed_objects = None  # type: ignore

from bluetooth_adapters import (
    DEFAULT_ADDRESS,
    AdapterDetails,
    AdvertisementHistory,
    DiscoveredDeviceAdvertisementData,
    DiscoveredDeviceAdvertisementDataDict,
    adapter_human_name,
    adapter_model,
    adapter_unique_name,
    discovered_device_advertisement_data_from_dict,
    discovered_device_advertisement_data_to_dict,
    expire_stale_scanner_discovered_device_advertisement_data,
    get_adapters,
    load_history_from_managed_objects,
)


@pytest.mark.asyncio
@pytest.mark.skipif(
    MessageType is None or get_dbus_managed_objects is None,
    reason="dbus_fast is not available",
)
async def test_get_bluetooth_adapters_file_not_found():
    """Test get_bluetooth_adapters()."""

    class MockMessageBus:
        def __init__(self, *args, **kwargs):
            raise FileNotFoundError

    with patch("bluetooth_adapters.dbus.MessageBus", MockMessageBus):
        assert await get_bluetooth_adapters() == []


@pytest.mark.asyncio
@pytest.mark.skipif(
    MessageType is None or get_dbus_managed_objects is None,
    reason="dbus_fast is not available",
)
async def test_get_bluetooth_adapters_connection_refused():
    """Test get_bluetooth_adapters with connection refused."""

    class MockMessageBus:
        def __init__(self, *args, **kwargs):
            raise ConnectionRefusedError

    with patch("bluetooth_adapters.dbus.MessageBus", MockMessageBus):
        assert await get_bluetooth_adapters() == []


@pytest.mark.asyncio
@pytest.mark.skipif(
    MessageType is None or get_dbus_managed_objects is None,
    reason="dbus_fast is not available",
)
async def test_get_bluetooth_adapters_auth_eror():
    """Test get_bluetooth_adapters with auth error."""

    class MockMessageBus:
        def __init__(self, *args, **kwargs):
            raise AuthError

    with patch("bluetooth_adapters.dbus.MessageBus", MockMessageBus):
        assert await get_bluetooth_adapters() == []


@pytest.mark.asyncio
@pytest.mark.skipif(
    MessageType is None or get_dbus_managed_objects is None,
    reason="dbus_fast is not available",
)
async def test_get_bluetooth_adapters_connect_refused_docker():
    class MockMessageBus:
        def __init__(self, *args, **kwargs):
            pass

        async def connect(self):
            raise ConnectionRefusedError

        def disconnect(self):
            pass

        async def call(self, *args, **kwargs):
            return None

    with (
        patch("bluetooth_adapters.dbus.MessageBus", MockMessageBus),
        patch("bluetooth_adapters.dbus.is_docker_env", return_value=True),
    ):
        assert await get_bluetooth_adapters() == []


@pytest.mark.asyncio
@pytest.mark.skipif(
    MessageType is None or get_dbus_managed_objects is None,
    reason="dbus_fast is not available",
)
async def test_get_bluetooth_adapters_connect_fails():
    class MockMessageBus:
        def __init__(self, *args, **kwargs):
            pass

        async def connect(self):
            raise FileNotFoundError

        def disconnect(self):
            pass

        async def call(self, *args, **kwargs):
            return None

    with patch("bluetooth_adapters.dbus.MessageBus", MockMessageBus):
        assert await get_bluetooth_adapters() == []


@pytest.mark.asyncio
@pytest.mark.skipif(
    MessageType is None or get_dbus_managed_objects is None,
    reason="dbus_fast is not available",
)
async def test_get_bluetooth_adapters_connect_fails_docker():
    class MockMessageBus:
        def __init__(self, *args, **kwargs):
            pass

        async def connect(self):
            raise FileNotFoundError

        def disconnect(self):
            pass

        async def call(self, *args, **kwargs):
            return None

    with (
        patch("bluetooth_adapters.dbus.MessageBus", MockMessageBus),
        patch("bluetooth_adapters.dbus.is_docker_env", return_value=True),
    ):
        assert await get_bluetooth_adapters() == []


@pytest.mark.asyncio
@pytest.mark.skipif(
    MessageType is None or get_dbus_managed_objects is None,
    reason="dbus_fast is not available",
)
async def test_get_bluetooth_adapters_connect_broken_pipe():
    class MockMessageBus:
        def __init__(self, *args, **kwargs):
            pass

        async def connect(self):
            raise BrokenPipeError

        def disconnect(self):
            pass

        async def call(self, *args, **kwargs):
            return None

    with patch("bluetooth_adapters.dbus.MessageBus", MockMessageBus):
        assert await get_bluetooth_adapters() == []


@pytest.mark.asyncio
@pytest.mark.skipif(
    MessageType is None or get_dbus_managed_objects is None,
    reason="dbus_fast is not available",
)
async def test_get_bluetooth_adapters_connect_broken_pipe_docker():
    class MockMessageBus:
        def __init__(self, *args, **kwargs):
            pass

        async def connect(self):
            raise BrokenPipeError

        def disconnect(self):
            pass

        async def call(self, *args, **kwargs):
            return None

    with (
        patch("bluetooth_adapters.dbus.MessageBus", MockMessageBus),
        patch("bluetooth_adapters.dbus.is_docker_env", return_value=True),
    ):
        assert await get_bluetooth_adapters() == []


@pytest.mark.asyncio
@pytest.mark.skipif(
    MessageType is None or get_dbus_managed_objects is None,
    reason="dbus_fast is not available",
)
async def test_get_bluetooth_adapters_connect_eof_error():
    class MockMessageBus:
        def __init__(self, *args, **kwargs):
            pass

        async def connect(self):
            pass

        def disconnect(self):
            pass

        async def call(self, *args, **kwargs):
            raise EOFError

    with patch("bluetooth_adapters.dbus.MessageBus", MockMessageBus):
        assert await get_bluetooth_adapters() == []


@pytest.mark.asyncio
@pytest.mark.skipif(
    MessageType is None or get_dbus_managed_objects is None,
    reason="dbus_fast is not available",
)
async def test_get_bluetooth_adapters_no_call_return():
    class MockMessageBus:
        def __init__(self, *args, **kwargs):
            pass

        async def connect(self):
            pass

        def disconnect(self):
            pass

        async def call(self, *args, **kwargs):
            return None

    with patch("bluetooth_adapters.dbus.MessageBus", MockMessageBus):
        assert await get_bluetooth_adapters() == []


@pytest.mark.asyncio
@pytest.mark.skipif(
    MessageType is None or get_dbus_managed_objects is None,
    reason="dbus_fast is not available",
)
async def test_get_bluetooth_adapters_times_out():
    async def _stall(*args: Any) -> None:
        await asyncio.sleep(10)

    class MockMessageBus:
        def __init__(self, *args, **kwargs):
            pass

        async def connect(self):
            pass

        async def call(self, *args, **kwargs):
            await _stall(*args)

        def disconnect(self):
            pass

    with (
        patch.object(bluetooth_adapters_dbus, "REPLY_TIMEOUT", 0),
        patch("bluetooth_adapters.dbus.MessageBus", MockMessageBus),
    ):
        assert await get_bluetooth_adapters() == []


@pytest.mark.asyncio
@pytest.mark.skipif(
    MessageType is None or get_dbus_managed_objects is None,
    reason="dbus_fast is not available",
)
async def test_get_bluetooth_adapters_no_wrong_return():
    class MockMessageBus:
        def __init__(self, *args, **kwargs):
            pass

        async def connect(self):
            pass

        def disconnect(self):
            pass

        async def call(self, *args, **kwargs):
            return MagicMock(
                body=[
                    {
                        "/org/bluez/hci0": "",
                        "/org/bluez/hci1": "",
                        "/org/bluez/hci1/any": "",
                    }
                ],
                message_type="wrong",
            )

    with patch("bluetooth_adapters.dbus.MessageBus", MockMessageBus):
        assert await get_bluetooth_adapters() == []


@pytest.mark.asyncio
@pytest.mark.skipif(
    MessageType is None or get_dbus_managed_objects is None,
    reason="dbus_fast is not available",
)
async def test_get_bluetooth_adapters_correct_return_valid_message():
    class MockMessageBus:
        def __init__(self, *args, **kwargs):
            pass

        async def connect(self):
            pass

        def disconnect(self):
            pass

        async def call(self, *args, **kwargs):
            return MagicMock(
                body=[
                    {
                        "/other": {},
                        "/org/bluez/hci0": {},
                        "/org/bluez/hci1": {},
                        "/org/bluez/hci1/any": {},
                    }
                ],
                message_type=MessageType.METHOD_RETURN,
            )

    with patch("bluetooth_adapters.dbus.MessageBus", MockMessageBus):
        assert await get_bluetooth_adapters() == ["hci0", "hci1"]


@pytest.mark.asyncio
@pytest.mark.skipif(
    MessageType is None or get_dbus_managed_objects is None,
    reason="dbus_fast is not available",
)
async def test_get_dbus_managed_objects():
    class MockMessageBus:
        def __init__(self, *args, **kwargs):
            pass

        async def connect(self):
            pass

        def disconnect(self):
            pass

        async def call(self, *args, **kwargs):
            return MagicMock(
                body=[
                    {
                        "/other": {},
                        "/org/bluez/hci0": {},
                        "/org/bluez/hci1": {},
                        "/org/bluez/hci1/any": {},
                    }
                ],
                message_type=MessageType.METHOD_RETURN,
            )

    with patch("bluetooth_adapters.dbus.MessageBus", MockMessageBus):
        assert await get_dbus_managed_objects() == {
            "/other": {},
            "/org/bluez/hci0": {},
            "/org/bluez/hci1": {},
            "/org/bluez/hci1/any": {},
        }


@pytest.mark.asyncio
@pytest.mark.skipif(
    MessageType is None or get_dbus_managed_objects is None,
    reason="dbus_fast is not available",
)
async def test_BlueZDBusObjects():
    class MockMessageBus:
        def __init__(self, *args, **kwargs):
            pass

        async def connect(self):
            pass

        def disconnect(self):
            pass

        async def call(self, *args, **kwargs):
            return MagicMock(
                body=[
                    {
                        "/other": {},
                        "/org/bluez/hci0": {},
                        "/org/bluez/hci1": {},
                        "/org/bluez/hci1/any": {},
                        "/org/bluez/hci0/dev_54_D2_72_AB_35_95": {
                            "org.freedesktop.DBus.Introspectable": {},
                            "org.bluez.Device1": {
                                "Address": "54:D2:72:AB:35:95",
                                "AddressType": "public",
                                "Name": "Nuki_1EAB3595",
                                "Alias": "Nuki_1EAB3595",
                                "Paired": False,
                                "Trusted": False,
                                "Blocked": False,
                                "LegacyPairing": False,
                                "RSSI": -78,
                                "Connected": False,
                                "UUIDs": [],
                                "Adapter": "/org/bluez/hci0",
                                "ManufacturerData": {
                                    "76": b"\\x02\\x15\\xa9.\\xe2\\x00U\\x01\\x11\\xe4\\x91l\\x08\\x00 \\x0c\\x9af\\x1e\\xab5\\x95\\xc4"
                                },
                                "ServicesResolved": False,
                                "AdvertisingFlags": {
                                    "__type": "<class 'bytearray'>",
                                    "repr": "bytearray(b'\\x06')",
                                },
                            },
                            "org.freedesktop.DBus.Properties": {},
                        },
                        "/org/bluez/hci1/dev_54_D2_72_AB_35_95": {
                            "org.freedesktop.DBus.Introspectable": {},
                            "org.bluez.Device1": {
                                "Address": "54:D2:72:AB:35:95",
                                "AddressType": "public",
                                "Name": "Nuki_1EAB3595",
                                "Alias": "Nuki_1EAB3595",
                                "Paired": False,
                                "Trusted": False,
                                "Blocked": False,
                                "LegacyPairing": False,
                                "RSSI": -100,
                                "Connected": False,
                                "UUIDs": [],
                                "Adapter": "/org/bluez/hci0",
                                "ManufacturerData": {
                                    "76": b"\\x02\\x15\\xa9.\\xe2\\x00U\\x01\\x11\\xe4\\x91l\\x08\\x00 \\x0c\\x9af\\x1e\\xab5\\x95\\xc4"
                                },
                                "ServicesResolved": False,
                                "AdvertisingFlags": {
                                    "__type": "<class 'bytearray'>",
                                    "repr": "bytearray(b'\\x06')",
                                },
                            },
                            "org.freedesktop.DBus.Properties": {},
                        },
                    }
                ],
                message_type=MessageType.METHOD_RETURN,
            )

    with patch("bluetooth_adapters.dbus.MessageBus", MockMessageBus):
        bluez = BlueZDBusObjects()
        await bluez.load()
        assert bluez.adapters == ["hci0", "hci1"]
        assert bluez.adapter_details == {"hci0": {}, "hci1": {}}
        assert bluez.history == {
            "54:D2:72:AB:35:95": AdvertisementHistory(ANY, ANY, "hci0")
        }
        assert bluez.history["54:D2:72:AB:35:95"].advertisement_data.rssi == -78
        assert (
            len(
                load_history_from_managed_objects(
                    bluez.unpacked_managed_objects, "hci1"
                )
            )
            == 1
        )


@pytest.mark.asyncio
@pytest.mark.skipif(
    MessageType is None or get_dbus_managed_objects is None,
    reason="dbus_fast is not available",
)
async def test_get_adapters_linux():
    """Test get_adapters."""

    class MockMessageBus:
        def __init__(self, *args, **kwargs):
            pass

        async def connect(self):
            pass

        def disconnect(self):
            pass

        async def call(self, *args, **kwargs):
            return MagicMock(
                body=[
                    {
                        "/other": {},
                        "/org/bluez/hci0": {
                            "org.bluez.Adapter1": {
                                "Address": "00:1A:7D:DA:71:04",
                                "AddressType": "public",
                                "Alias": "homeassistant",
                                "Class": 2883584,
                                "Discoverable": False,
                                "DiscoverableTimeout": 180,
                                "Discovering": True,
                                "Modalias": "usb:v1D6Bp0246d053F",
                                "Name": "homeassistant",
                                "Pairable": False,
                                "PairableTimeout": 0,
                                "Powered": True,
                                "Roles": ["central", "peripheral"],
                                "UUIDs": [
                                    "0000110e-0000-1000-8000-00805f9b34fb",
                                    "0000110a-0000-1000-8000-00805f9b34fb",
                                    "00001200-0000-1000-8000-00805f9b34fb",
                                    "0000110b-0000-1000-8000-00805f9b34fb",
                                    "00001108-0000-1000-8000-00805f9b34fb",
                                    "0000110c-0000-1000-8000-00805f9b34fb",
                                    "00001800-0000-1000-8000-00805f9b34fb",
                                    "00001801-0000-1000-8000-00805f9b34fb",
                                    "0000180a-0000-1000-8000-00805f9b34fb",
                                    "00001112-0000-1000-8000-00805f9b34fb",
                                ],
                            },
                            "org.bluez.GattManager1": {},
                            "org.bluez.LEAdvertisingManager1": {
                                "ActiveInstances": 0,
                                "SupportedIncludes": [
                                    "tx-power",
                                    "appearance",
                                    "local-name",
                                ],
                                "SupportedInstances": 5,
                            },
                            "org.bluez.Media1": {},
                            "org.bluez.NetworkServer1": {},
                            "org.freedesktop.DBus.Introspectable": {},
                            "org.freedesktop.DBus.Properties": {},
                        },
                        "/org/bluez/hci1": {},
                        "/org/bluez/hci2": {
                            "org.bluez.Adapter1": {
                                "Address": "00:00:00:00:00:00",
                                "AddressType": "public",
                                "Alias": "homeassistant",
                                "Class": 2883584,
                                "Discoverable": False,
                                "DiscoverableTimeout": 180,
                                "Discovering": True,
                                "Modalias": "usb:v1D6Bp0246d053F",
                                "Name": "homeassistant",
                                "Pairable": False,
                                "PairableTimeout": 0,
                                "Powered": True,
                                "Roles": ["central", "peripheral"],
                                "UUIDs": [
                                    "0000110e-0000-1000-8000-00805f9b34fb",
                                    "0000110a-0000-1000-8000-00805f9b34fb",
                                    "00001200-0000-1000-8000-00805f9b34fb",
                                    "0000110b-0000-1000-8000-00805f9b34fb",
                                    "00001108-0000-1000-8000-00805f9b34fb",
                                    "0000110c-0000-1000-8000-00805f9b34fb",
                                    "00001800-0000-1000-8000-00805f9b34fb",
                                    "00001801-0000-1000-8000-00805f9b34fb",
                                    "0000180a-0000-1000-8000-00805f9b34fb",
                                    "00001112-0000-1000-8000-00805f9b34fb",
                                ],
                            },
                            "org.bluez.GattManager1": {},
                            "org.bluez.LEAdvertisingManager1": {
                                "ActiveInstances": 0,
                                "SupportedIncludes": [
                                    "tx-power",
                                    "appearance",
                                    "local-name",
                                ],
                                "SupportedInstances": 5,
                            },
                            "org.bluez.Media1": {},
                            "org.bluez.NetworkServer1": {},
                            "org.freedesktop.DBus.Introspectable": {},
                            "org.freedesktop.DBus.Properties": {},
                        },
                        "/org/bluez/hci3": {
                            "org.bluez.Adapter1": {
                                "Address": "00:1A:7D:DA:71:05",
                                "AddressType": "public",
                                "Alias": "homeassistant",
                                "Class": 2883584,
                                "Discoverable": False,
                                "DiscoverableTimeout": 180,
                                "Discovering": True,
                                "Modalias": "usb:v1D6Bp0246d053F",
                                "Name": "homeassistant",
                                "Pairable": False,
                                "PairableTimeout": 0,
                                "Powered": True,
                                "Roles": ["central", "peripheral"],
                                "UUIDs": [
                                    "0000110e-0000-1000-8000-00805f9b34fb",
                                    "0000110a-0000-1000-8000-00805f9b34fb",
                                    "00001200-0000-1000-8000-00805f9b34fb",
                                    "0000110b-0000-1000-8000-00805f9b34fb",
                                    "00001108-0000-1000-8000-00805f9b34fb",
                                    "0000110c-0000-1000-8000-00805f9b34fb",
                                    "00001800-0000-1000-8000-00805f9b34fb",
                                    "00001801-0000-1000-8000-00805f9b34fb",
                                    "0000180a-0000-1000-8000-00805f9b34fb",
                                    "00001112-0000-1000-8000-00805f9b34fb",
                                ],
                            },
                            "org.bluez.GattManager1": {},
                            "org.bluez.LEAdvertisingManager1": {
                                "ActiveInstances": 0,
                                "SupportedIncludes": [
                                    "tx-power",
                                    "appearance",
                                    "local-name",
                                ],
                                "SupportedInstances": 5,
                            },
                            "org.bluez.Media1": {},
                            "org.bluez.NetworkServer1": {},
                            "org.freedesktop.DBus.Introspectable": {},
                            "org.freedesktop.DBus.Properties": {},
                        },
                        "/org/bluez/hci1/any": {},
                        "/org/bluez/hci0/dev_54_D2_72_AB_35_95": {
                            "org.freedesktop.DBus.Introspectable": {},
                            "org.bluez.Device1": {
                                "Address": "54:D2:72:AB:35:95",
                                "AddressType": "public",
                                "Name": "Nuki_1EAB3595",
                                "Alias": "Nuki_1EAB3595",
                                "Paired": False,
                                "Trusted": False,
                                "Blocked": False,
                                "LegacyPairing": False,
                                "RSSI": -78,
                                "Connected": False,
                                "UUIDs": [],
                                "Adapter": "/org/bluez/hci0",
                                "ManufacturerData": {
                                    "76": b"\\x02\\x15\\xa9.\\xe2\\x00U\\x01\\x11\\xe4\\x91l\\x08\\x00 \\x0c\\x9af\\x1e\\xab5\\x95\\xc4"
                                },
                                "ServicesResolved": False,
                                "AdvertisingFlags": {
                                    "__type": "<class 'bytearray'>",
                                    "repr": "bytearray(b'\\x06')",
                                },
                            },
                            "org.freedesktop.DBus.Properties": {},
                        },
                        "/org/bluez/hci1/dev_54_D2_72_AB_35_95": {
                            "org.freedesktop.DBus.Introspectable": {},
                            "org.bluez.Device1": {
                                "Address": "54:D2:72:AB:35:95",
                                "AddressType": "public",
                                "Name": "Nuki_1EAB3595",
                                "Alias": "Nuki_1EAB3595",
                                "Paired": False,
                                "Trusted": False,
                                "Blocked": False,
                                "LegacyPairing": False,
                                "RSSI": -100,
                                "Connected": False,
                                "UUIDs": [],
                                "Adapter": "/org/bluez/hci0",
                                "ManufacturerData": {
                                    "76": b"\\x02\\x15\\xa9.\\xe2\\x00U\\x01\\x11\\xe4\\x91l\\x08\\x00 \\x0c\\x9af\\x1e\\xab5\\x95\\xc4"
                                },
                                "ServicesResolved": False,
                                "AdvertisingFlags": {
                                    "__type": "<class 'bytearray'>",
                                    "repr": "bytearray(b'\\x06')",
                                },
                            },
                            "org.freedesktop.DBus.Properties": {},
                        },
                    }
                ],
                message_type=MessageType.METHOD_RETURN,
            )

    class MockUSBDevice(USBDevice):
        def __init__(self, *args, **kwargs):
            self.manufacturer = "XTech"
            self.product = "Bluetooth 4.0 USB Adapter"
            self.vendor_id = "0a12"
            self.product_id = "0001"
            pass

    class MockBluetoothDevice(USBBluetoothDevice):
        def __init__(self, *args, **kwargs):
            self.usb_device = MockUSBDevice()
            pass

        def setup(self, *args, **kwargs):
            pass

    with (
        patch("platform.system", return_value="Linux"),
        patch("platform.release", return_value="18.7.0"),
        patch("bluetooth_adapters.dbus.MessageBus", MockMessageBus),
        patch(
            "bluetooth_adapters.systems.linux.get_adapters_from_hci", return_value={}
        ),
        patch(
            "bluetooth_adapters.systems.linux.USBBluetoothDevice", MockBluetoothDevice
        ),
    ):
        bluetooth_adapters = get_adapters()
        await bluetooth_adapters.refresh()
        assert bluetooth_adapters.default_adapter == "hci0"
        assert bluetooth_adapters.history == {
            "54:D2:72:AB:35:95": AdvertisementHistory(
                device=ANY, advertisement_data=ANY, source="hci0"
            )
        }
        # hci0 should show
        # hci1 is empty so it should not be in the list
        # hci2 should not show as 00:00:00:00:00:00 are filtered downstream now
        # hci3 should show
        assert bluetooth_adapters.adapters == {
            "hci0": {
                "address": "00:1A:7D:DA:71:04",
                "hw_version": "usb:v1D6Bp0246d053F",
                "manufacturer": "XTech",
                "passive_scan": False,
                "product": "Bluetooth 4.0 USB Adapter",
                "product_id": "0001",
                "sw_version": "18.7.0",
                "vendor_id": "0a12",
                "adapter_type": "usb",
                "advertise": True,
                "powered": True,
            },
            "hci2": {
                "address": "00:00:00:00:00:00",
                "hw_version": "usb:v1D6Bp0246d053F",
                "manufacturer": "XTech",
                "passive_scan": False,
                "product": "Bluetooth 4.0 USB Adapter",
                "product_id": "0001",
                "sw_version": "18.7.0",
                "vendor_id": "0a12",
                "adapter_type": "usb",
                "advertise": True,
                "powered": True,
            },
            "hci3": {
                "address": "00:1A:7D:DA:71:05",
                "hw_version": "usb:v1D6Bp0246d053F",
                "manufacturer": "XTech",
                "passive_scan": False,
                "product": "Bluetooth 4.0 USB Adapter",
                "product_id": "0001",
                "sw_version": "18.7.0",
                "vendor_id": "0a12",
                "adapter_type": "usb",
                "advertise": True,
                "powered": True,
            },
        }


@pytest.mark.asyncio
@pytest.mark.skipif(
    MessageType is None or get_dbus_managed_objects is None,
    reason="dbus_fast is not available",
)
async def test_get_adapters_linux_device_listed_before_adapter():
    """
    Test get_adapters. List a device entry before the adapter entry to ensure
    the adapter is retrieved and not the device.
    """

    class MockMessageBus:
        def __init__(self, *args, **kwargs):
            pass

        async def connect(self):
            pass

        def disconnect(self):
            pass

        async def call(self, *args, **kwargs):
            return MagicMock(
                body=[
                    {
                        "/other": {},
                        "/org/bluez/hci0/dev_54_D2_72_AB_35_95": {
                            "org.freedesktop.DBus.Introspectable": {},
                            "org.bluez.Device1": {
                                "Address": "54:D2:72:AB:35:95",
                                "AddressType": "public",
                                "Name": "Nuki_1EAB3595",
                                "Alias": "Nuki_1EAB3595",
                                "Paired": False,
                                "Trusted": False,
                                "Blocked": False,
                                "LegacyPairing": False,
                                "RSSI": -78,
                                "Connected": False,
                                "UUIDs": [],
                                "Adapter": "/org/bluez/hci0",
                                "ManufacturerData": {
                                    "76": b"\\x02\\x15\\xa9.\\xe2\\x00U\\x01\\x11\\xe4\\x91l\\x08\\x00 \\x0c\\x9af\\x1e\\xab5\\x95\\xc4"
                                },
                                "ServicesResolved": False,
                                "AdvertisingFlags": {
                                    "__type": "<class 'bytearray'>",
                                    "repr": "bytearray(b'\\x06')",
                                },
                            },
                            "org.freedesktop.DBus.Properties": {},
                        },
                        "/org/bluez/hci0": {
                            "org.bluez.Adapter1": {
                                "Address": "00:1A:7D:DA:71:04",
                                "AddressType": "public",
                                "Alias": "homeassistant",
                                "Class": 2883584,
                                "Discoverable": False,
                                "DiscoverableTimeout": 180,
                                "Discovering": True,
                                "Modalias": "usb:v1D6Bp0246d053F",
                                "Name": "homeassistant",
                                "Pairable": False,
                                "PairableTimeout": 0,
                                "Powered": True,
                                "Roles": ["central", "peripheral"],
                                "UUIDs": [
                                    "0000110e-0000-1000-8000-00805f9b34fb",
                                    "0000110a-0000-1000-8000-00805f9b34fb",
                                    "00001200-0000-1000-8000-00805f9b34fb",
                                    "0000110b-0000-1000-8000-00805f9b34fb",
                                    "00001108-0000-1000-8000-00805f9b34fb",
                                    "0000110c-0000-1000-8000-00805f9b34fb",
                                    "00001800-0000-1000-8000-00805f9b34fb",
                                    "00001801-0000-1000-8000-00805f9b34fb",
                                    "0000180a-0000-1000-8000-00805f9b34fb",
                                    "00001112-0000-1000-8000-00805f9b34fb",
                                ],
                            },
                            "org.bluez.GattManager1": {},
                            "org.bluez.LEAdvertisingManager1": {
                                "ActiveInstances": 0,
                                "SupportedIncludes": [
                                    "tx-power",
                                    "appearance",
                                    "local-name",
                                ],
                                "SupportedInstances": 5,
                            },
                            "org.bluez.Media1": {},
                            "org.bluez.NetworkServer1": {},
                            "org.freedesktop.DBus.Introspectable": {},
                            "org.freedesktop.DBus.Properties": {},
                        },
                        "/org/bluez/hci1": {},
                        "/org/bluez/hci2": {
                            "org.bluez.Adapter1": {
                                "Address": "00:00:00:00:00:00",
                                "AddressType": "public",
                                "Alias": "homeassistant",
                                "Class": 2883584,
                                "Discoverable": False,
                                "DiscoverableTimeout": 180,
                                "Discovering": True,
                                "Modalias": "usb:v1D6Bp0246d053F",
                                "Name": "homeassistant",
                                "Pairable": False,
                                "PairableTimeout": 0,
                                "Powered": True,
                                "Roles": ["central", "peripheral"],
                                "UUIDs": [
                                    "0000110e-0000-1000-8000-00805f9b34fb",
                                    "0000110a-0000-1000-8000-00805f9b34fb",
                                    "00001200-0000-1000-8000-00805f9b34fb",
                                    "0000110b-0000-1000-8000-00805f9b34fb",
                                    "00001108-0000-1000-8000-00805f9b34fb",
                                    "0000110c-0000-1000-8000-00805f9b34fb",
                                    "00001800-0000-1000-8000-00805f9b34fb",
                                    "00001801-0000-1000-8000-00805f9b34fb",
                                    "0000180a-0000-1000-8000-00805f9b34fb",
                                    "00001112-0000-1000-8000-00805f9b34fb",
                                ],
                            },
                            "org.bluez.GattManager1": {},
                            "org.bluez.LEAdvertisingManager1": {
                                "ActiveInstances": 0,
                                "SupportedIncludes": [
                                    "tx-power",
                                    "appearance",
                                    "local-name",
                                ],
                                "SupportedInstances": 5,
                            },
                            "org.bluez.Media1": {},
                            "org.bluez.NetworkServer1": {},
                            "org.freedesktop.DBus.Introspectable": {},
                            "org.freedesktop.DBus.Properties": {},
                        },
                        "/org/bluez/hci3": {
                            "org.bluez.Adapter1": {
                                "Address": "00:1A:7D:DA:71:05",
                                "AddressType": "public",
                                "Alias": "homeassistant",
                                "Class": 2883584,
                                "Discoverable": False,
                                "DiscoverableTimeout": 180,
                                "Discovering": True,
                                "Modalias": "usb:v1D6Bp0246d053F",
                                "Name": "homeassistant",
                                "Pairable": False,
                                "PairableTimeout": 0,
                                "Powered": True,
                                "Roles": ["central", "peripheral"],
                                "UUIDs": [
                                    "0000110e-0000-1000-8000-00805f9b34fb",
                                    "0000110a-0000-1000-8000-00805f9b34fb",
                                    "00001200-0000-1000-8000-00805f9b34fb",
                                    "0000110b-0000-1000-8000-00805f9b34fb",
                                    "00001108-0000-1000-8000-00805f9b34fb",
                                    "0000110c-0000-1000-8000-00805f9b34fb",
                                    "00001800-0000-1000-8000-00805f9b34fb",
                                    "00001801-0000-1000-8000-00805f9b34fb",
                                    "0000180a-0000-1000-8000-00805f9b34fb",
                                    "00001112-0000-1000-8000-00805f9b34fb",
                                ],
                            },
                            "org.bluez.GattManager1": {},
                            "org.bluez.LEAdvertisingManager1": {
                                "ActiveInstances": 0,
                                "SupportedIncludes": [
                                    "tx-power",
                                    "appearance",
                                    "local-name",
                                ],
                                "SupportedInstances": 5,
                            },
                            "org.bluez.Media1": {},
                            "org.bluez.NetworkServer1": {},
                            "org.freedesktop.DBus.Introspectable": {},
                            "org.freedesktop.DBus.Properties": {},
                        },
                        "/org/bluez/hci1/any": {},
                        "/org/bluez/hci1/dev_54_D2_72_AB_35_95": {
                            "org.freedesktop.DBus.Introspectable": {},
                            "org.bluez.Device1": {
                                "Address": "54:D2:72:AB:35:95",
                                "AddressType": "public",
                                "Name": "Nuki_1EAB3595",
                                "Alias": "Nuki_1EAB3595",
                                "Paired": False,
                                "Trusted": False,
                                "Blocked": False,
                                "LegacyPairing": False,
                                "RSSI": -100,
                                "Connected": False,
                                "UUIDs": [],
                                "Adapter": "/org/bluez/hci0",
                                "ManufacturerData": {
                                    "76": b"\\x02\\x15\\xa9.\\xe2\\x00U\\x01\\x11\\xe4\\x91l\\x08\\x00 \\x0c\\x9af\\x1e\\xab5\\x95\\xc4"
                                },
                                "ServicesResolved": False,
                                "AdvertisingFlags": {
                                    "__type": "<class 'bytearray'>",
                                    "repr": "bytearray(b'\\x06')",
                                },
                            },
                            "org.freedesktop.DBus.Properties": {},
                        },
                    }
                ],
                message_type=MessageType.METHOD_RETURN,
            )

    class MockUSBDevice(USBDevice):
        def __init__(self, *args, **kwargs):
            self.manufacturer = "XTech"
            self.product = "Bluetooth 4.0 USB Adapter"
            self.vendor_id = "0a12"
            self.product_id = "0001"
            pass

    class MockBluetoothDevice(USBBluetoothDevice):
        def __init__(self, *args, **kwargs):
            self.usb_device = MockUSBDevice()
            pass

        def setup(self, *args, **kwargs):
            pass

    with (
        patch("platform.system", return_value="Linux"),
        patch("platform.release", return_value="18.7.0"),
        patch("bluetooth_adapters.dbus.MessageBus", MockMessageBus),
        patch(
            "bluetooth_adapters.systems.linux.get_adapters_from_hci", return_value={}
        ),
        patch(
            "bluetooth_adapters.systems.linux.USBBluetoothDevice", MockBluetoothDevice
        ),
    ):
        bluetooth_adapters = get_adapters()
        await bluetooth_adapters.refresh()
        assert bluetooth_adapters.default_adapter == "hci0"
        assert bluetooth_adapters.history == {
            "54:D2:72:AB:35:95": AdvertisementHistory(
                device=ANY, advertisement_data=ANY, source="hci0"
            )
        }
        # hci0 should show
        # hci1 is empty so it should not be in the list
        # hci2 should not show as 00:00:00:00:00:00 are filtered downstream now
        # hci3 should show
        assert bluetooth_adapters.adapters == {
            "hci0": {
                "address": "00:1A:7D:DA:71:04",
                "hw_version": "usb:v1D6Bp0246d053F",
                "manufacturer": "XTech",
                "passive_scan": False,
                "product": "Bluetooth 4.0 USB Adapter",
                "product_id": "0001",
                "sw_version": "18.7.0",
                "vendor_id": "0a12",
                "adapter_type": "usb",
                "advertise": True,
                "powered": True,
            },
            "hci2": {
                "address": "00:00:00:00:00:00",
                "hw_version": "usb:v1D6Bp0246d053F",
                "manufacturer": "XTech",
                "passive_scan": False,
                "product": "Bluetooth 4.0 USB Adapter",
                "product_id": "0001",
                "sw_version": "18.7.0",
                "vendor_id": "0a12",
                "adapter_type": "usb",
                "advertise": True,
                "powered": True,
            },
            "hci3": {
                "address": "00:1A:7D:DA:71:05",
                "hw_version": "usb:v1D6Bp0246d053F",
                "manufacturer": "XTech",
                "passive_scan": False,
                "product": "Bluetooth 4.0 USB Adapter",
                "product_id": "0001",
                "sw_version": "18.7.0",
                "vendor_id": "0a12",
                "adapter_type": "usb",
                "advertise": True,
                "powered": True,
            },
        }


@pytest.mark.asyncio
@pytest.mark.skipif(
    MessageType is None or get_dbus_managed_objects is None,
    reason="dbus_fast is not available",
)
async def test_get_adapters_linux_uart():
    """Test get_adapters with uart devices."""

    class MockMessageBus:
        def __init__(self, *args, **kwargs):
            pass

        async def connect(self):
            pass

        def disconnect(self):
            pass

        async def call(self, *args, **kwargs):
            return MagicMock(
                body=[
                    {
                        "/other": {},
                        "/org/bluez/hci0": {
                            "org.bluez.Adapter1": {
                                "Address": "00:1A:7D:DA:71:04",
                                "AddressType": "public",
                                "Alias": "homeassistant",
                                "Class": 2883584,
                                "Discoverable": False,
                                "DiscoverableTimeout": 180,
                                "Discovering": True,
                                "Modalias": "usb:v1D6Bp0246d053F",
                                "Name": "homeassistant",
                                "Pairable": False,
                                "PairableTimeout": 0,
                                "Powered": True,
                                "Roles": ["central", "peripheral"],
                                "UUIDs": [
                                    "0000110e-0000-1000-8000-00805f9b34fb",
                                    "0000110a-0000-1000-8000-00805f9b34fb",
                                    "00001200-0000-1000-8000-00805f9b34fb",
                                    "0000110b-0000-1000-8000-00805f9b34fb",
                                    "00001108-0000-1000-8000-00805f9b34fb",
                                    "0000110c-0000-1000-8000-00805f9b34fb",
                                    "00001800-0000-1000-8000-00805f9b34fb",
                                    "00001801-0000-1000-8000-00805f9b34fb",
                                    "0000180a-0000-1000-8000-00805f9b34fb",
                                    "00001112-0000-1000-8000-00805f9b34fb",
                                ],
                            },
                            "org.bluez.GattManager1": {},
                            "org.bluez.LEAdvertisingManager1": {
                                "ActiveInstances": 0,
                                "SupportedIncludes": [
                                    "tx-power",
                                    "appearance",
                                    "local-name",
                                ],
                                "SupportedInstances": 5,
                            },
                            "org.bluez.Media1": {},
                            "org.bluez.NetworkServer1": {},
                            "org.freedesktop.DBus.Introspectable": {},
                            "org.freedesktop.DBus.Properties": {},
                        },
                        "/org/bluez/hci1": {},
                        "/org/bluez/hci2": {
                            "org.bluez.Adapter1": {
                                "Address": "00:00:00:00:00:00",
                                "AddressType": "public",
                                "Alias": "homeassistant",
                                "Class": 2883584,
                                "Discoverable": False,
                                "DiscoverableTimeout": 180,
                                "Discovering": True,
                                "Modalias": "usb:v1D6Bp0246d053F",
                                "Name": "homeassistant",
                                "Pairable": False,
                                "PairableTimeout": 0,
                                "Powered": True,
                                "Roles": ["central", "peripheral"],
                                "UUIDs": [
                                    "0000110e-0000-1000-8000-00805f9b34fb",
                                    "0000110a-0000-1000-8000-00805f9b34fb",
                                    "00001200-0000-1000-8000-00805f9b34fb",
                                    "0000110b-0000-1000-8000-00805f9b34fb",
                                    "00001108-0000-1000-8000-00805f9b34fb",
                                    "0000110c-0000-1000-8000-00805f9b34fb",
                                    "00001800-0000-1000-8000-00805f9b34fb",
                                    "00001801-0000-1000-8000-00805f9b34fb",
                                    "0000180a-0000-1000-8000-00805f9b34fb",
                                    "00001112-0000-1000-8000-00805f9b34fb",
                                ],
                            },
                            "org.bluez.GattManager1": {},
                            "org.bluez.LEAdvertisingManager1": {
                                "ActiveInstances": 0,
                                "SupportedIncludes": [
                                    "tx-power",
                                    "appearance",
                                    "local-name",
                                ],
                                "SupportedInstances": 5,
                            },
                            "org.bluez.Media1": {},
                            "org.bluez.NetworkServer1": {},
                            "org.freedesktop.DBus.Introspectable": {},
                            "org.freedesktop.DBus.Properties": {},
                        },
                        "/org/bluez/hci3": {
                            "org.bluez.Adapter1": {
                                "Address": "00:1A:7D:DA:71:05",
                                "AddressType": "public",
                                "Alias": "homeassistant",
                                "Class": 2883584,
                                "Discoverable": False,
                                "DiscoverableTimeout": 180,
                                "Discovering": True,
                                "Modalias": "usb:v1D6Bp0246d053F",
                                "Name": "homeassistant",
                                "Pairable": False,
                                "PairableTimeout": 0,
                                "Powered": True,
                                "Roles": ["central", "peripheral"],
                                "UUIDs": [
                                    "0000110e-0000-1000-8000-00805f9b34fb",
                                    "0000110a-0000-1000-8000-00805f9b34fb",
                                    "00001200-0000-1000-8000-00805f9b34fb",
                                    "0000110b-0000-1000-8000-00805f9b34fb",
                                    "00001108-0000-1000-8000-00805f9b34fb",
                                    "0000110c-0000-1000-8000-00805f9b34fb",
                                    "00001800-0000-1000-8000-00805f9b34fb",
                                    "00001801-0000-1000-8000-00805f9b34fb",
                                    "0000180a-0000-1000-8000-00805f9b34fb",
                                    "00001112-0000-1000-8000-00805f9b34fb",
                                ],
                            },
                            "org.bluez.GattManager1": {},
                            "org.bluez.LEAdvertisingManager1": {
                                "ActiveInstances": 0,
                                "SupportedIncludes": [
                                    "tx-power",
                                    "appearance",
                                    "local-name",
                                ],
                                "SupportedInstances": 5,
                            },
                            "org.bluez.Media1": {},
                            "org.bluez.NetworkServer1": {},
                            "org.freedesktop.DBus.Introspectable": {},
                            "org.freedesktop.DBus.Properties": {},
                        },
                        "/org/bluez/hci1/any": {},
                        "/org/bluez/hci0/dev_54_D2_72_AB_35_95": {
                            "org.freedesktop.DBus.Introspectable": {},
                            "org.bluez.Device1": {
                                "Address": "54:D2:72:AB:35:95",
                                "AddressType": "public",
                                "Name": "Nuki_1EAB3595",
                                "Alias": "Nuki_1EAB3595",
                                "Paired": False,
                                "Trusted": False,
                                "Blocked": False,
                                "LegacyPairing": False,
                                "RSSI": -78,
                                "Connected": False,
                                "UUIDs": [],
                                "Adapter": "/org/bluez/hci0",
                                "ManufacturerData": {
                                    "76": b"\\x02\\x15\\xa9.\\xe2\\x00U\\x01\\x11\\xe4\\x91l\\x08\\x00 \\x0c\\x9af\\x1e\\xab5\\x95\\xc4"
                                },
                                "ServicesResolved": False,
                                "AdvertisingFlags": {
                                    "__type": "<class 'bytearray'>",
                                    "repr": "bytearray(b'\\x06')",
                                },
                            },
                            "org.freedesktop.DBus.Properties": {},
                        },
                        "/org/bluez/hci1/dev_54_D2_72_AB_35_95": {
                            "org.freedesktop.DBus.Introspectable": {},
                            "org.bluez.Device1": {
                                "Address": "54:D2:72:AB:35:95",
                                "AddressType": "public",
                                "Name": "Nuki_1EAB3595",
                                "Alias": "Nuki_1EAB3595",
                                "Paired": False,
                                "Trusted": False,
                                "Blocked": False,
                                "LegacyPairing": False,
                                "RSSI": -100,
                                "Connected": False,
                                "UUIDs": [],
                                "Adapter": "/org/bluez/hci0",
                                "ManufacturerData": {
                                    "76": b"\\x02\\x15\\xa9.\\xe2\\x00U\\x01\\x11\\xe4\\x91l\\x08\\x00 \\x0c\\x9af\\x1e\\xab5\\x95\\xc4"
                                },
                                "ServicesResolved": False,
                                "AdvertisingFlags": {
                                    "__type": "<class 'bytearray'>",
                                    "repr": "bytearray(b'\\x06')",
                                },
                            },
                            "org.freedesktop.DBus.Properties": {},
                        },
                    }
                ],
                message_type=MessageType.METHOD_RETURN,
            )

    class MockUARTDevice(UARTDevice):
        def __init__(self, *args, **kwargs):
            self.manufacturer = "XTech"
            self.product = "Bluetooth 4.0 USB Adapter"
            pass

    class MockUARTBluetoothDevice(UARTBluetoothDevice):
        def __init__(self, *args, **kwargs):
            self.uart_device = MockUARTDevice()
            pass

        def setup(self, *args, **kwargs):
            pass

    class MockUSBBluetoothDevice(UARTBluetoothDevice):
        def __init__(self, *args, **kwargs):
            self.uart_device = MockUARTDevice()
            pass

        def setup(self, *args, **kwargs):
            raise NotAUSBDeviceError

    with (
        patch("platform.system", return_value="Linux"),
        patch("platform.release", return_value="18.7.0"),
        patch("bluetooth_adapters.dbus.MessageBus", MockMessageBus),
        patch(
            "bluetooth_adapters.systems.linux.get_adapters_from_hci", return_value={}
        ),
        patch(
            "bluetooth_adapters.systems.linux.USBBluetoothDevice",
            MockUSBBluetoothDevice,
        ),
        patch(
            "bluetooth_adapters.systems.linux.UARTBluetoothDevice",
            MockUARTBluetoothDevice,
        ),
    ):
        bluetooth_adapters = get_adapters()
        await bluetooth_adapters.refresh()
        assert bluetooth_adapters.default_adapter == "hci0"
        assert bluetooth_adapters.history == {
            "54:D2:72:AB:35:95": AdvertisementHistory(
                device=ANY, advertisement_data=ANY, source="hci0"
            )
        }
        # hci0 should show
        # hci1 is empty so it should not be in the list
        # hci2 should not show as 00:00:00:00:00:00 are filtered downstream now
        # hci3 should show
        assert bluetooth_adapters.adapters == {
            "hci0": {
                "address": "00:1A:7D:DA:71:04",
                "hw_version": "usb:v1D6Bp0246d053F",
                "manufacturer": "cyber-blue(HK)Ltd",
                "passive_scan": False,
                "product": "Bluetooth 4.0 USB Adapter",
                "product_id": None,
                "sw_version": "18.7.0",
                "vendor_id": None,
                "adapter_type": "uart",
                "advertise": True,
                "powered": True,
            },
            "hci2": {
                "address": "00:00:00:00:00:00",
                "hw_version": "usb:v1D6Bp0246d053F",
                "manufacturer": "XTech",
                "passive_scan": False,
                "product": "Bluetooth 4.0 USB Adapter",
                "product_id": None,
                "sw_version": "18.7.0",
                "vendor_id": None,
                "adapter_type": "uart",
                "advertise": True,
                "powered": True,
            },
            "hci3": {
                "address": "00:1A:7D:DA:71:05",
                "hw_version": "usb:v1D6Bp0246d053F",
                "manufacturer": "cyber-blue(HK)Ltd",
                "passive_scan": False,
                "product": "Bluetooth 4.0 USB Adapter",
                "product_id": None,
                "sw_version": "18.7.0",
                "vendor_id": None,
                "adapter_type": "uart",
                "advertise": True,
                "powered": True,
            },
        }


@pytest.mark.asyncio
@pytest.mark.skipif(
    MessageType is None or get_dbus_managed_objects is None,
    reason="dbus_fast is not available",
)
async def test_get_adapters_linux_no_usb_device():
    """Test get_adapters."""

    class MockMessageBus:
        def __init__(self, *args, **kwargs):
            pass

        async def connect(self):
            pass

        def disconnect(self):
            pass

        async def call(self, *args, **kwargs):
            return MagicMock(
                body=[
                    {
                        "/other": {},
                        "/org/bluez/hci3": {
                            "org.bluez.Adapter1": {
                                "Address": "00:1A:7D:DA:71:04",
                                "AddressType": "public",
                                "Alias": "homeassistant",
                                "Class": 2883584,
                                "Discoverable": False,
                                "DiscoverableTimeout": 180,
                                "Discovering": True,
                                "Modalias": "usb:v1D6Bp0246d053F",
                                "Name": "homeassistant",
                                "Pairable": False,
                                "PairableTimeout": 0,
                                "Powered": True,
                                "Roles": ["central", "peripheral"],
                                "UUIDs": [
                                    "0000110e-0000-1000-8000-00805f9b34fb",
                                    "0000110a-0000-1000-8000-00805f9b34fb",
                                    "00001200-0000-1000-8000-00805f9b34fb",
                                    "0000110b-0000-1000-8000-00805f9b34fb",
                                    "00001108-0000-1000-8000-00805f9b34fb",
                                    "0000110c-0000-1000-8000-00805f9b34fb",
                                    "00001800-0000-1000-8000-00805f9b34fb",
                                    "00001801-0000-1000-8000-00805f9b34fb",
                                    "0000180a-0000-1000-8000-00805f9b34fb",
                                    "00001112-0000-1000-8000-00805f9b34fb",
                                ],
                            },
                            "org.bluez.GattManager1": {},
                            "org.bluez.LEAdvertisingManager1": {
                                "ActiveInstances": 0,
                                "SupportedIncludes": [
                                    "tx-power",
                                    "appearance",
                                    "local-name",
                                ],
                                "SupportedInstances": 5,
                            },
                            "org.bluez.Media1": {},
                            "org.bluez.NetworkServer1": {},
                            "org.freedesktop.DBus.Introspectable": {},
                            "org.freedesktop.DBus.Properties": {},
                        },
                        "/org/bluez/hci4": {},
                        "/org/bluez/hci5/any": {},
                        "/org/bluez/hci3/dev_54_D2_72_AB_35_95": {
                            "org.freedesktop.DBus.Introspectable": {},
                            "org.bluez.Device1": {
                                "Address": "54:D2:72:AB:35:95",
                                "AddressType": "public",
                                "Name": "Nuki_1EAB3595",
                                "Alias": "Nuki_1EAB3595",
                                "Paired": False,
                                "Trusted": False,
                                "Blocked": False,
                                "LegacyPairing": False,
                                "RSSI": -78,
                                "Connected": False,
                                "UUIDs": [],
                                "Adapter": "/org/bluez/hci3",
                                "ManufacturerData": {
                                    "76": b"\\x02\\x15\\xa9.\\xe2\\x00U\\x01\\x11\\xe4\\x91l\\x08\\x00 \\x0c\\x9af\\x1e\\xab5\\x95\\xc4"
                                },
                                "ServicesResolved": False,
                                "AdvertisingFlags": {
                                    "__type": "<class 'bytearray'>",
                                    "repr": "bytearray(b'\\x06')",
                                },
                            },
                            "org.freedesktop.DBus.Properties": {},
                        },
                        "/org/bluez/hci1/dev_54_D2_72_AB_35_95": {
                            "org.freedesktop.DBus.Introspectable": {},
                            "org.bluez.Device1": {
                                "Address": "54:D2:72:AB:35:95",
                                "AddressType": "public",
                                "Name": "Nuki_1EAB3595",
                                "Alias": "Nuki_1EAB3595",
                                "Paired": False,
                                "Trusted": False,
                                "Blocked": False,
                                "LegacyPairing": False,
                                "RSSI": -100,
                                "Connected": False,
                                "UUIDs": [],
                                "Adapter": "/org/bluez/hci0",
                                "ManufacturerData": {
                                    "76": b"\\x02\\x15\\xa9.\\xe2\\x00U\\x01\\x11\\xe4\\x91l\\x08\\x00 \\x0c\\x9af\\x1e\\xab5\\x95\\xc4"
                                },
                                "ServicesResolved": False,
                                "AdvertisingFlags": {
                                    "__type": "<class 'bytearray'>",
                                    "repr": "bytearray(b'\\x06')",
                                },
                            },
                            "org.freedesktop.DBus.Properties": {},
                        },
                    }
                ],
                message_type=MessageType.METHOD_RETURN,
            )

    class NoMfrMockUSBDevice(USBDevice):
        def __init__(self, *args, **kwargs):
            self.manufacturer = None
            self.product = "Bluetooth 4.0 USB Adapter"
            self.vendor_id = "0a12"
            self.product_id = "0001"
            pass

    class NoMfrMockBluetoothDevice(USBBluetoothDevice):
        def __init__(self, *args, **kwargs):
            self.usb_device = NoMfrMockUSBDevice()
            pass

        def setup(self, *args, **kwargs):
            pass

    with (
        patch("platform.system", return_value="Linux"),
        patch("platform.release", return_value="18.7.0"),
        patch("bluetooth_adapters.dbus.MessageBus", MockMessageBus),
        patch(
            "bluetooth_adapters.systems.linux.get_adapters_from_hci", return_value={}
        ),
        patch(
            "bluetooth_adapters.systems.linux.USBBluetoothDevice",
            NoMfrMockBluetoothDevice,
        ),
    ):
        bluetooth_adapters = get_adapters()
        await bluetooth_adapters.refresh()
        assert bluetooth_adapters.default_adapter == "hci0"
        assert bluetooth_adapters.history == {
            "54:D2:72:AB:35:95": AdvertisementHistory(
                device=ANY, advertisement_data=ANY, source="hci3"
            )
        }
        assert bluetooth_adapters.adapters == {
            "hci3": {
                "address": "00:1A:7D:DA:71:04",
                "manufacturer": "cyber-blue(HK)Ltd",
                "product": "Bluetooth 4.0 USB Adapter",
                "vendor_id": "0a12",
                "product_id": "0001",
                "hw_version": "usb:v1D6Bp0246d053F",
                "passive_scan": False,
                "sw_version": "18.7.0",
                "adapter_type": "usb",
                "advertise": True,
                "powered": True,
            },
        }


@pytest.mark.asyncio
@pytest.mark.skipif(
    MessageType is None or get_dbus_managed_objects is None,
    reason="dbus_fast is not available",
)
async def test_get_adapters_linux_hci_only_adapter():
    """Test adapters reported only by HCI (not by BlueZ) are merged in.

    Exercises the HCI branch of ``LinuxAdapters.adapters``: an adapter known to
    the kernel via HCI but absent from BlueZ should appear with the MAC-derived
    manufacturer and ``adapter_type`` of ``None``. The empty-MAC adapter should
    have a ``None`` manufacturer (no vendor lookup attempted).
    """

    class MockMessageBus:
        def __init__(self, *args, **kwargs):
            pass

        async def connect(self):
            pass

        def disconnect(self):
            pass

        async def call(self, *args, **kwargs):
            # BlueZ reports no adapters; everything comes from HCI.
            return MagicMock(body=[{}], message_type=MessageType.METHOD_RETURN)

    with (
        patch("platform.system", return_value="Linux"),
        patch("platform.release", return_value="18.7.0"),
        patch("bluetooth_adapters.dbus.MessageBus", MockMessageBus),
        patch(
            "bluetooth_adapters.systems.linux.get_adapters_from_hci",
            return_value={
                5: {
                    "name": "hci5",
                    "bdaddr": "00:1A:7D:DA:71:13",
                    "powered": True,
                    "advertise": True,
                },
                9: {
                    "name": "hci9",
                    "bdaddr": "00:00:00:00:00:00",
                    "powered": False,
                    "advertise": False,
                },
            },
        ),
        patch(
            "bluetooth_adapters.systems.linux.aiooui.get_vendor",
            return_value="Acme Bluetooth",
        ),
    ):
        bluetooth_adapters = get_adapters()
        await bluetooth_adapters.refresh()
        assert bluetooth_adapters.adapters == {
            "hci5": {
                "address": "00:1A:7D:DA:71:13",
                "sw_version": "18.7.0",
                "hw_version": None,
                "passive_scan": False,
                "manufacturer": "Acme Bluetooth",
                "product": None,
                "vendor_id": None,
                "product_id": None,
                "adapter_type": None,
                "powered": True,
                "advertise": True,
            },
            "hci9": {
                "address": "00:00:00:00:00:00",
                "sw_version": "18.7.0",
                "hw_version": None,
                "passive_scan": False,
                "manufacturer": None,
                "product": None,
                "vendor_id": None,
                "product_id": None,
                "adapter_type": None,
                "powered": False,
                "advertise": False,
            },
        }


@pytest.mark.asyncio
async def test_get_adapters_macos():
    """Test get_adapters macos."""
    with (
        patch("platform.system", return_value="Darwin"),
        patch("platform.release", return_value="18.7.0"),
    ):
        bluetooth_adapters = get_adapters()
        await bluetooth_adapters.refresh()
        assert bluetooth_adapters.default_adapter == "Core Bluetooth"
        assert bluetooth_adapters.history == {}
        assert bluetooth_adapters.adapters == {
            "Core Bluetooth": {
                "address": "00:00:00:00:00:00",
                "passive_scan": False,
                "sw_version": "18.7.0",
                "manufacturer": "Apple",
                "product": "Unknown MacOS Model",
                "vendor_id": "Unknown",
                "product_id": "Unknown",
                "adapter_type": None,
                "advertise": True,
                "powered": True,
            }
        }


@pytest.mark.asyncio
async def test_get_adapters_windows():
    """Test get_adapters windows."""
    with (
        patch("platform.system", return_value="Windows"),
        patch("platform.release", return_value="18.7.0"),
    ):
        bluetooth_adapters = get_adapters()
        await bluetooth_adapters.refresh()
        assert bluetooth_adapters.default_adapter == "bluetooth"
        assert bluetooth_adapters.history == {}
        assert bluetooth_adapters.adapters == {
            "bluetooth": {
                "address": "00:00:00:00:00:00",
                "passive_scan": False,
                "sw_version": "18.7.0",
                "manufacturer": "Microsoft",
                "product": "Unknown Windows Model",
                "vendor_id": "Unknown",
                "product_id": "Unknown",
                "adapter_type": None,
                "advertise": True,
                "powered": True,
            }
        }


def test_adapter_human_name():
    """Test adapter human name."""
    assert adapter_human_name("hci0", DEFAULT_ADDRESS) == "hci0"
    assert adapter_human_name("hci0", "aa:bb:cc:dd:ee:ff") == "hci0 (aa:bb:cc:dd:ee:ff)"


@pytest.mark.asyncio
async def test_get_manufacturer_from_mac():
    """Test get_manufacturer_from_mac."""
    assert await get_manufacturer_from_mac("00:00:00:00:00:00") is None
    assert await get_manufacturer_from_mac("00:1A:7D:DA:71:04") == "cyber-blue(HK)Ltd"


def test_adapter_unique_name():
    """Test adapter unique name."""
    assert adapter_unique_name("hci0", DEFAULT_ADDRESS) == "hci0"
    assert adapter_unique_name("hci0", "aa:bb:cc:dd:ee:ff") == "aa:bb:cc:dd:ee:ff"


def test_adapter_model():
    """Test adapter model."""
    windows_details = AdapterDetails(
        {
            "address": "00:00:00:00:00:00",
            "passive_scan": False,
            "sw_version": "18.7.0",
            "manufacturer": "Microsoft",
            "product": "Unknown Windows Model",
            "vendor_id": "Unknown",
            "product_id": "Unknown",
            "advertise": True,
            "powered": True,
        }
    )
    assert adapter_model(windows_details) == "Unknown Windows Model"
    linux_details = AdapterDetails(
        {
            "address": "00:1A:7D:DA:71:04",
            "manufacturer": "XTech",
            "product": "Bluetooth 4.0 USB Adapter",
            "vendor_id": "0a12",
            "product_id": "0001",
            "hw_version": "usb:v1D6Bp0246d053F",
            "passive_scan": False,
            "sw_version": "18.7.0",
            "advertise": True,
            "powered": True,
        }
    )
    assert adapter_model(linux_details) == "Bluetooth 4.0 USB Adapter (0a12:0001)"


def test_discovered_device_advertisement_data_to_dict():
    """Test discovered device advertisement data to dict."""
    result = discovered_device_advertisement_data_to_dict(
        DiscoveredDeviceAdvertisementData(
            True,
            100,
            {
                "AA:BB:CC:DD:EE:FF": (
                    BLEDevice(
                        address="AA:BB:CC:DD:EE:FF",
                        name="Test Device",
                        details={"details": "test"},
                    ),
                    AdvertisementData(
                        local_name="Test Device",
                        manufacturer_data={0x004C: b"\x02\x15\xaa\xbb\xcc\xdd\xee\xff"},
                        tx_power=50,
                        service_data={
                            "0000180d-0000-1000-8000-00805f9b34fb": b"\x00\x00\x00\x00"
                        },
                        service_uuids=["0000180d-0000-1000-8000-00805f9b34fb"],
                        platform_data=("Test Device", ""),
                        rssi=-50,
                    ),
                )
            },
            {"AA:BB:CC:DD:EE:FF": 100000},
        )
    )
    assert result == {
        "connectable": True,
        "discovered_device_advertisement_datas": {
            "AA:BB:CC:DD:EE:FF": {
                "advertisement_data": {
                    "local_name": "Test Device",
                    "manufacturer_data": {"76": "0215aabbccddeeff"},
                    "rssi": -50,
                    "service_data": {
                        "0000180d-0000-1000-8000-00805f9b34fb": "00000000"
                    },
                    "service_uuids": ["0000180d-0000-1000-8000-00805f9b34fb"],
                    "tx_power": 50,
                    "platform_data": ["Test Device", ""],
                },
                "device": {
                    "address": "AA:BB:CC:DD:EE:FF",
                    "details": {"details": "test"},
                    "name": "Test Device",
                    "rssi": -50,
                },
            }
        },
        "discovered_device_timestamps": {"AA:BB:CC:DD:EE:FF": ANY},
        "expire_seconds": 100,
    }


def test_discovered_device_advertisement_data_from_dict():
    now = time.time()
    result = discovered_device_advertisement_data_from_dict(
        {
            "connectable": True,
            "discovered_device_advertisement_datas": {
                "AA:BB:CC:DD:EE:FF": {
                    "advertisement_data": {
                        "local_name": "Test Device",
                        "manufacturer_data": {"76": "0215aabbccddeeff"},
                        "rssi": -50,
                        "service_data": {
                            "0000180d-0000-1000-8000-00805f9b34fb": "00000000"
                        },
                        "service_uuids": ["0000180d-0000-1000-8000-00805f9b34fb"],
                        "tx_power": 50,
                        "platform_data": ["Test Device", ""],
                    },
                    "device": {
                        "address": "AA:BB:CC:DD:EE:FF",
                        "details": {"details": "test"},
                        "name": "Test Device",
                        "rssi": -50,
                    },
                }
            },
            "discovered_device_timestamps": {"AA:BB:CC:DD:EE:FF": now},
            "expire_seconds": 100,
        }
    )

    expected_ble_device = BLEDevice(
        address="AA:BB:CC:DD:EE:FF",
        name="Test Device",
        details={"details": "test"},
    )

    expected_advertisement_data = AdvertisementData(
        local_name="Test Device",
        manufacturer_data={0x004C: b"\x02\x15\xaa\xbb\xcc\xdd\xee\xff"},
        tx_power=50,
        service_data={"0000180d-0000-1000-8000-00805f9b34fb": b"\x00\x00\x00\x00"},
        service_uuids=["0000180d-0000-1000-8000-00805f9b34fb"],
        platform_data=("Test Device", ""),
        rssi=-50,
    )
    assert result is not None
    out_ble_device = result.discovered_device_advertisement_datas["AA:BB:CC:DD:EE:FF"][
        0
    ]
    out_advertisement_data = result.discovered_device_advertisement_datas[
        "AA:BB:CC:DD:EE:FF"
    ][1]
    assert out_ble_device.address == expected_ble_device.address
    assert out_ble_device.name == expected_ble_device.name
    assert out_ble_device.details == expected_ble_device.details
    assert out_advertisement_data == expected_advertisement_data

    assert result == DiscoveredDeviceAdvertisementData(
        connectable=True,
        expire_seconds=100,
        discovered_device_advertisement_datas={
            "AA:BB:CC:DD:EE:FF": (
                ANY,
                expected_advertisement_data,
            )
        },
        discovered_device_timestamps={"AA:BB:CC:DD:EE:FF": ANY},
    )


@pytest.mark.skipif(
    MessageType is None or get_dbus_managed_objects is None,
    reason="dbus_fast is not available",
)
def test_expire_stale_scanner_discovered_device_advertisement_data():
    """Test expire_stale_scanner_discovered_device_advertisement_data."""
    now = time.time()
    data = {
        "myscanner": DiscoveredDeviceAdvertisementDataDict(
            {
                "connectable": True,
                "discovered_device_advertisement_datas": {
                    "AA:BB:CC:DD:EE:FF": {
                        "advertisement_data": {
                            "local_name": "Test Device",
                            "manufacturer_data": {"76": "0215aabbccddeeff"},
                            "rssi": -50,
                            "service_data": {
                                "0000180d-0000-1000-8000-00805f9b34fb": "00000000"
                            },
                            "service_uuids": ["0000180d-0000-1000-8000-00805f9b34fb"],
                            "tx_power": 50,
                            "platform_data": ["Test Device", ""],
                        },
                        "device": {
                            "address": "AA:BB:CC:DD:EE:FF",
                            "details": {"details": "test"},
                            "name": "Test Device",
                            "rssi": -50,
                        },
                    },
                    "CC:DD:EE:FF:AA:BB": {
                        "advertisement_data": {
                            "local_name": "Test Device Expired",
                            "manufacturer_data": {"76": "0215aabbccddeeff"},
                            "rssi": -50,
                            "service_data": {
                                "0000180d-0000-1000-8000-00805f9b34fb": "00000000"
                            },
                            "service_uuids": ["0000180d-0000-1000-8000-00805f9b34fb"],
                            "tx_power": 50,
                            "platform_data": ["Test Device", ""],
                        },
                        "device": {
                            "address": "CC:DD:EE:FF:AA:BB",
                            "details": {"details": "test"},
                            "name": "Test Device Expired",
                            "rssi": -50,
                        },
                    },
                },
                "discovered_device_timestamps": {
                    "AA:BB:CC:DD:EE:FF": now,
                    "CC:DD:EE:FF:AA:BB": now - 100,
                },
                "expire_seconds": 100,
            }
        ),
        "all_expired": DiscoveredDeviceAdvertisementDataDict(
            {
                "connectable": True,
                "discovered_device_advertisement_datas": {
                    "CC:DD:EE:FF:AA:BB": {
                        "advertisement_data": {
                            "local_name": "Test Device Expired",
                            "manufacturer_data": {"76": "0215aabbccddeeff"},
                            "rssi": -50,
                            "service_data": {
                                "0000180d-0000-1000-8000-00805f9b34fb": "00000000"
                            },
                            "service_uuids": ["0000180d-0000-1000-8000-00805f9b34fb"],
                            "tx_power": 50,
                            "platform_data": ["Test Device", ""],
                        },
                        "device": {
                            "address": "CC:DD:EE:FF:AA:BB",
                            "details": {"details": "test"},
                            "name": "Test Device Expired",
                            "rssi": -50,
                        },
                    }
                },
                "discovered_device_timestamps": {"CC:DD:EE:FF:AA:BB": now - 100},
                "expire_seconds": 100,
            }
        ),
    }
    expire_stale_scanner_discovered_device_advertisement_data(data)
    assert len(data["myscanner"]["discovered_device_advertisement_datas"]) == 1
    assert (
        "CC:DD:EE:FF:AA:BB"
        not in data["myscanner"]["discovered_device_advertisement_datas"]
    )
    assert "all_expired" not in data


@pytest.mark.skipif(
    MessageType is None or get_dbus_managed_objects is None,
    reason="dbus_fast is not available",
)
def test_expire_future_discovered_device_advertisement_data(
    caplog: pytest.LogCaptureFixture,
) -> None:
    """Test test_expire_future_discovered_device_advertisement_data."""
    now = time.time()
    data = {
        "myscanner": DiscoveredDeviceAdvertisementDataDict(
            {
                "connectable": True,
                "discovered_device_advertisement_datas": {
                    "AA:BB:CC:DD:EE:FF": {
                        "advertisement_data": {
                            "local_name": "Test Device",
                            "manufacturer_data": {"76": "0215aabbccddeeff"},
                            "rssi": -50,
                            "service_data": {
                                "0000180d-0000-1000-8000-00805f9b34fb": "00000000"
                            },
                            "service_uuids": ["0000180d-0000-1000-8000-00805f9b34fb"],
                            "tx_power": 50,
                            "platform_data": ["Test Device", ""],
                        },
                        "device": {
                            "address": "AA:BB:CC:DD:EE:FF",
                            "details": {"details": "test"},
                            "name": "Test Device",
                            "rssi": -50,
                        },
                    },
                    "CC:DD:EE:FF:AA:BB": {
                        "advertisement_data": {
                            "local_name": "Test Device Expired",
                            "manufacturer_data": {"76": "0215aabbccddeeff"},
                            "rssi": -50,
                            "service_data": {
                                "0000180d-0000-1000-8000-00805f9b34fb": "00000000"
                            },
                            "service_uuids": ["0000180d-0000-1000-8000-00805f9b34fb"],
                            "tx_power": 50,
                            "platform_data": ["Test Device", ""],
                        },
                        "device": {
                            "address": "CC:DD:EE:FF:AA:BB",
                            "details": {"details": "test"},
                            "name": "Test Device Expired",
                            "rssi": -50,
                        },
                    },
                },
                "discovered_device_timestamps": {
                    "AA:BB:CC:DD:EE:FF": now,
                    "CC:DD:EE:FF:AA:BB": now - 100,
                },
                "expire_seconds": 100,
            }
        ),
        "all_future": DiscoveredDeviceAdvertisementDataDict(
            {
                "connectable": True,
                "discovered_device_advertisement_datas": {
                    "CC:DD:EE:FF:AA:BB": {
                        "advertisement_data": {
                            "local_name": "Test Device Expired",
                            "manufacturer_data": {"76": "0215aabbccddeeff"},
                            "rssi": -50,
                            "service_data": {
                                "0000180d-0000-1000-8000-00805f9b34fb": "00000000"
                            },
                            "service_uuids": ["0000180d-0000-1000-8000-00805f9b34fb"],
                            "tx_power": 50,
                            "platform_data": ["Test Device", ""],
                        },
                        "device": {
                            "address": "CC:DD:EE:FF:AA:BB",
                            "details": {"details": "test"},
                            "name": "Test Device Expired",
                            "rssi": -50,
                        },
                    }
                },
                "discovered_device_timestamps": {"CC:DD:EE:FF:AA:BB": now + 1000000},
                "expire_seconds": 100,
            }
        ),
    }
    expire_stale_scanner_discovered_device_advertisement_data(data)
    assert len(data["myscanner"]["discovered_device_advertisement_datas"]) == 1
    assert (
        "CC:DD:EE:FF:AA:BB"
        not in data["myscanner"]["discovered_device_advertisement_datas"]
    )
    assert "all_future" not in data
    assert (
        "for CC:DD:EE:FF:AA:BB on scanner all_future as it is the future" in caplog.text
    )


def test_discovered_device_advertisement_data_from_dict_corrupt(caplog):
    """Test discovered_device_advertisement_data_from_dict with corrupt data."""
    now = time.time()
    result = discovered_device_advertisement_data_from_dict(
        {
            "connectable": True,
            "discovered_device_advertisement_datas": {
                "AA:BB:CC:DD:EE:FF": {
                    "advertisement_data": {  # type: ignore[typeddict-item]
                        "local_name": "Test Device",
                        "manufacturer_data": {"76": "0215aabbccddeeff"},
                        "rssi": -50,
                        "service_data": {
                            "0000180d-0000-1000-8000-00805f9b34fb": "00000000"
                        },
                        "service_uuids": ["0000180d-0000-1000-8000-00805f9b34fb"],
                    },
                    "device": {  # type: ignore[typeddict-item]
                        "address": "AA:BB:CC:DD:EE:FF",
                        "details": {"details": "test"},
                        "rssi": -50,
                    },
                }
            },
            "discovered_device_timestamps": {"AA:BB:CC:DD:EE:FF": now},
            "expire_seconds": 100,
        }
    )
    assert result is None
    assert "Error deserializing discovered_device_advertisement_data" in caplog.text


def _single_adapter_message_bus(address, *, modalias="usb:v1D6Bp0246d053F"):
    """Build a MockMessageBus exposing a single BlueZ adapter at hci0.

    Keeps the manufacturer-resolution branch tests compact: only the
    properties read by ``LinuxAdapters.adapters`` are populated.
    """

    body = {
        "/org/bluez/hci0": {
            "org.bluez.Adapter1": {
                "Address": address,
                "Powered": True,
                "Modalias": modalias,
            },
            "org.bluez.LEAdvertisingManager1": {},
        },
    }

    class MockMessageBus:
        def __init__(self, *args, **kwargs):
            pass

        async def connect(self):
            pass

        def disconnect(self):
            pass

        async def call(self, *args, **kwargs):
            return MagicMock(body=[body], message_type=MessageType.METHOD_RETURN)

    return MockMessageBus


@pytest.mark.asyncio
@pytest.mark.skipif(
    MessageType is None or get_dbus_managed_objects is None,
    reason="dbus_fast is not available",
)
async def test_get_adapters_linux_caches_result():
    """adapters is computed once and the same object is returned on reaccess."""

    class MockUSBDevice(USBDevice):
        def __init__(self, *args, **kwargs):
            self.manufacturer = "XTech"
            self.product = "Bluetooth 4.0 USB Adapter"
            self.vendor_id = "0a12"
            self.product_id = "0001"

    class MockBluetoothDevice(USBBluetoothDevice):
        def __init__(self, *args, **kwargs):
            self.usb_device = MockUSBDevice()

        def setup(self, *args, **kwargs):
            pass

    with (
        patch("platform.system", return_value="Linux"),
        patch("platform.release", return_value="18.7.0"),
        patch(
            "bluetooth_adapters.dbus.MessageBus",
            _single_adapter_message_bus("00:1A:7D:DA:71:04"),
        ),
        patch(
            "bluetooth_adapters.systems.linux.get_adapters_from_hci", return_value={}
        ),
        patch(
            "bluetooth_adapters.systems.linux.USBBluetoothDevice", MockBluetoothDevice
        ),
    ):
        adapters = get_adapters()
        await adapters.refresh()
        first = adapters.adapters
        # Second access must return the cached object, not recompute it.
        assert adapters.adapters is first
        # A device manufacturer that is real (not the vendor id / Unknown / None)
        # is used verbatim instead of the MAC vendor lookup.
        assert first["hci0"]["manufacturer"] == "XTech"
        assert first["hci0"]["adapter_type"] == "usb"


@pytest.mark.asyncio
@pytest.mark.skipif(
    MessageType is None or get_dbus_managed_objects is None,
    reason="dbus_fast is not available",
)
async def test_get_adapters_linux_usb_device_missing():
    """A USB adapter whose usb_device is None and address is empty resolves blank."""

    class MockBluetoothDevice(USBBluetoothDevice):
        def __init__(self, *args, **kwargs):
            self.usb_device = None

        def setup(self, *args, **kwargs):
            pass

    with (
        patch("platform.system", return_value="Linux"),
        patch("platform.release", return_value="18.7.0"),
        patch(
            "bluetooth_adapters.dbus.MessageBus",
            _single_adapter_message_bus(DEFAULT_ADDRESS),
        ),
        patch(
            "bluetooth_adapters.systems.linux.get_adapters_from_hci", return_value={}
        ),
        patch(
            "bluetooth_adapters.systems.linux.USBBluetoothDevice", MockBluetoothDevice
        ),
    ):
        adapters = get_adapters()
        await adapters.refresh()
        assert adapters.adapters == {
            "hci0": {
                "address": DEFAULT_ADDRESS,
                "hw_version": "usb:v1D6Bp0246d053F",
                "manufacturer": None,
                "passive_scan": False,
                "product": None,
                "product_id": None,
                "sw_version": "18.7.0",
                "vendor_id": None,
                "adapter_type": "usb",
                "advertise": True,
                "powered": True,
            },
        }


@pytest.mark.asyncio
@pytest.mark.skipif(
    MessageType is None or get_dbus_managed_objects is None,
    reason="dbus_fast is not available",
)
async def test_get_adapters_linux_uart_device_missing():
    """A UART adapter whose uart_device is None falls back to the MAC vendor."""

    class MockUSB(USBBluetoothDevice):
        def __init__(self, *args, **kwargs):
            self.usb_device = None

        def setup(self, *args, **kwargs):
            raise NotAUSBDeviceError

    class MockUART(UARTBluetoothDevice):
        def __init__(self, *args, **kwargs):
            self.uart_device = None

        def setup(self, *args, **kwargs):
            pass

    with (
        patch("platform.system", return_value="Linux"),
        patch("platform.release", return_value="18.7.0"),
        patch(
            "bluetooth_adapters.dbus.MessageBus",
            _single_adapter_message_bus("00:1A:7D:DA:71:04"),
        ),
        patch(
            "bluetooth_adapters.systems.linux.get_adapters_from_hci", return_value={}
        ),
        patch("bluetooth_adapters.systems.linux.USBBluetoothDevice", MockUSB),
        patch("bluetooth_adapters.systems.linux.UARTBluetoothDevice", MockUART),
    ):
        adapters = get_adapters()
        await adapters.refresh()
        assert adapters.adapters["hci0"]["adapter_type"] == "uart"
        assert adapters.adapters["hci0"]["manufacturer"] == "cyber-blue(HK)Ltd"
        assert adapters.adapters["hci0"]["product"] is None


@pytest.mark.asyncio
@pytest.mark.skipif(
    MessageType is None or get_dbus_managed_objects is None,
    reason="dbus_fast is not available",
)
async def test_get_adapters_linux_uart_device_missing_empty_address():
    """A UART adapter with no uart_device and an empty address has no vendor."""

    class MockUSB(USBBluetoothDevice):
        def __init__(self, *args, **kwargs):
            self.usb_device = None

        def setup(self, *args, **kwargs):
            raise NotAUSBDeviceError

    class MockUART(UARTBluetoothDevice):
        def __init__(self, *args, **kwargs):
            self.uart_device = None

        def setup(self, *args, **kwargs):
            pass

    with (
        patch("platform.system", return_value="Linux"),
        patch("platform.release", return_value="18.7.0"),
        patch(
            "bluetooth_adapters.dbus.MessageBus",
            _single_adapter_message_bus(DEFAULT_ADDRESS),
        ),
        patch(
            "bluetooth_adapters.systems.linux.get_adapters_from_hci", return_value={}
        ),
        patch("bluetooth_adapters.systems.linux.USBBluetoothDevice", MockUSB),
        patch("bluetooth_adapters.systems.linux.UARTBluetoothDevice", MockUART),
    ):
        adapters = get_adapters()
        await adapters.refresh()
        assert adapters.adapters["hci0"]["adapter_type"] == "uart"
        # Empty address skips the MAC vendor lookup; nothing to resolve from.
        assert adapters.adapters["hci0"]["manufacturer"] is None


@pytest.mark.asyncio
@pytest.mark.skipif(
    MessageType is None or get_dbus_managed_objects is None,
    reason="dbus_fast is not available",
)
async def test_refresh_devices_setup_file_not_found():
    """A FileNotFoundError during setup stops probing; the adapter stays USB."""

    class MockUSB(USBBluetoothDevice):
        def __init__(self, *args, **kwargs):
            self.usb_device = None

        def setup(self, *args, **kwargs):
            raise FileNotFoundError

    class MockUART(UARTBluetoothDevice):
        def __init__(self, *args, **kwargs):
            self.uart_device = None

        def setup(self, *args, **kwargs):
            pass

    with (
        patch("platform.system", return_value="Linux"),
        patch("platform.release", return_value="18.7.0"),
        patch(
            "bluetooth_adapters.dbus.MessageBus",
            _single_adapter_message_bus("00:1A:7D:DA:71:04"),
        ),
        patch(
            "bluetooth_adapters.systems.linux.get_adapters_from_hci", return_value={}
        ),
        patch("bluetooth_adapters.systems.linux.USBBluetoothDevice", MockUSB),
        patch("bluetooth_adapters.systems.linux.UARTBluetoothDevice", MockUART),
    ):
        adapters = get_adapters()
        await adapters.refresh()
        # break (not continue) means we never fell through to the UART probe.
        assert adapters.adapters["hci0"]["adapter_type"] == "usb"


@pytest.mark.asyncio
@pytest.mark.skipif(
    MessageType is None or get_dbus_managed_objects is None,
    reason="dbus_fast is not available",
)
async def test_refresh_devices_setup_unexpected_error(caplog):
    """An unexpected setup error is logged and stops probing for that adapter."""

    class MockUSB(USBBluetoothDevice):
        def __init__(self, *args, **kwargs):
            self.usb_device = None

        def setup(self, *args, **kwargs):
            raise ValueError("boom")

    class MockUART(UARTBluetoothDevice):
        def __init__(self, *args, **kwargs):
            self.uart_device = None

        def setup(self, *args, **kwargs):
            pass

    with (
        patch("platform.system", return_value="Linux"),
        patch("platform.release", return_value="18.7.0"),
        patch(
            "bluetooth_adapters.dbus.MessageBus",
            _single_adapter_message_bus("00:1A:7D:DA:71:04"),
        ),
        patch(
            "bluetooth_adapters.systems.linux.get_adapters_from_hci", return_value={}
        ),
        patch("bluetooth_adapters.systems.linux.USBBluetoothDevice", MockUSB),
        patch("bluetooth_adapters.systems.linux.UARTBluetoothDevice", MockUART),
    ):
        adapters = get_adapters()
        await adapters.refresh()
        assert adapters.adapters["hci0"]["adapter_type"] == "usb"
    assert "Unexpected error setting up device" in caplog.text


@pytest.mark.asyncio
@pytest.mark.skipif(
    MessageType is None or get_dbus_managed_objects is None,
    reason="dbus_fast is not available",
)
async def test_refresh_devices_neither_usb_nor_uart():
    """When both probes reject the device, the last-tried (UART) type wins."""

    class MockUSB(USBBluetoothDevice):
        def __init__(self, *args, **kwargs):
            self.usb_device = None

        def setup(self, *args, **kwargs):
            raise NotAUSBDeviceError

    class MockUART(UARTBluetoothDevice):
        def __init__(self, *args, **kwargs):
            self.uart_device = None

        def setup(self, *args, **kwargs):
            raise NotAUARTDeviceError

    with (
        patch("platform.system", return_value="Linux"),
        patch("platform.release", return_value="18.7.0"),
        patch(
            "bluetooth_adapters.dbus.MessageBus",
            _single_adapter_message_bus("00:1A:7D:DA:71:04"),
        ),
        patch(
            "bluetooth_adapters.systems.linux.get_adapters_from_hci", return_value={}
        ),
        patch("bluetooth_adapters.systems.linux.USBBluetoothDevice", MockUSB),
        patch("bluetooth_adapters.systems.linux.UARTBluetoothDevice", MockUART),
    ):
        adapters = get_adapters()
        await adapters.refresh()
        # Both probes raised; the adapter still surfaces with MAC-derived vendor.
        assert adapters.adapters["hci0"]["adapter_type"] == "uart"
        assert adapters.adapters["hci0"]["manufacturer"] == "cyber-blue(HK)Ltd"



from __future__ import annotations

import platform

from ..adapters import BluetoothAdapters


def get_adapters() -> BluetoothAdapters:
    """Get the adapters."""
    if platform.system() == "Windows":
        from .windows import WindowsAdapters

        return WindowsAdapters()
    if platform.system() == "Darwin":
        from .macos import MacOSAdapters

        return MacOSAdapters()
    from .linux import LinuxAdapters

    return LinuxAdapters()



"""Get manufacturer from mac address."""

from __future__ import annotations

import aiooui

from .const import EMPTY_MAC_ADDRESS


async def get_manufacturer_from_mac(mac_address: str) -> str | None:
    """Get manufacturer from mac address."""
    if not aiooui.is_loaded():
        await aiooui.async_load()
    if mac_address == EMPTY_MAC_ADDRESS:
        return None
    return aiooui.get_vendor(mac_address)



"""Models for bluetooth adapters."""

from __future__ import annotations

from typing import Final, TypedDict


class AdapterDetails(TypedDict, total=False):
    """Adapter details."""

    address: str
    sw_version: str
    hw_version: str | None
    manufacturer: str | None
    product: str | None
    vendor_id: str | None
    product_id: str | None
    passive_scan: bool
    connection_slots: int | None
    adapter_type: str | None  # "usb", "uart", or None
    advertise: bool
    """Whether the adapter supports advertising."""

    powered: bool
    """Whether the adapter is powered on."""


ADAPTER_ADDRESS: Final = "address"
ADAPTER_SW_VERSION: Final = "sw_version"
ADAPTER_HW_VERSION: Final = "hw_version"
ADAPTER_PASSIVE_SCAN: Final = "passive_scan"
ADAPTER_MANUFACTURER: Final = "manufacturer"
ADAPTER_PRODUCT: Final = "product"
ADAPTER_VENDOR_ID: Final = "vendor_id"
ADAPTER_PRODUCT_ID: Final = "product_id"
ADAPTER_CONNECTION_SLOTS: Final = "connection_slots"
ADAPTER_TYPE: Final = "adapter_type"
ADAPTER_ADVERTISE: Final = "advertise"
ADAPTER_POWERED: Final = "powered"



"""Tests for the BluetoothAdapters abstract base class."""

import pytest
from bluetooth_adapters.adapters import BluetoothAdapters
from bluetooth_adapters.models import AdapterDetails


class CompleteAdapters(BluetoothAdapters):
    """A subclass that implements every abstract member."""

    @property
    def adapters(self):
        return {"hci0": AdapterDetails(address="00:11:22:33:44:55")}

    @property
    def default_adapter(self):
        return "hci0"


class MissingDefaultAdapter(BluetoothAdapters):
    """A subclass that forgets to implement default_adapter."""

    @property
    def adapters(self):
        return {}


def test_base_class_cannot_be_instantiated():
    """The base class is abstract and must not be instantiable."""
    with pytest.raises(TypeError):
        BluetoothAdapters()  # type: ignore[abstract]


def test_incomplete_subclass_cannot_be_instantiated():
    """A subclass missing an abstract member stays abstract."""
    with pytest.raises(TypeError):
        MissingDefaultAdapter()  # type: ignore[abstract]


def test_complete_subclass_can_be_instantiated():
    """A subclass implementing every abstract member instantiates cleanly."""
    adapters = CompleteAdapters()
    assert adapters.adapters == {"hci0": AdapterDetails(address="00:11:22:33:44:55")}
    assert adapters.default_adapter == "hci0"


@pytest.mark.asyncio
async def test_concrete_defaults_on_subclass():
    """refresh() and history have usable concrete defaults from the base."""
    adapters = CompleteAdapters()
    await adapters.refresh()
    assert adapters.history == {}



from __future__ import annotations

from dataclasses import dataclass
from typing import Any

from bleak.backends.device import BLEDevice
from bleak.backends.scanner import AdvertisementData

MIN_RSSI = -127


@dataclass
class AdvertisementHistory:
    device: BLEDevice
    advertisement_data: AdvertisementData
    source: str


def load_history_from_managed_objects(
    managed_objects: dict[str, Any], source_adapter: str | None = None
) -> dict[str, AdvertisementHistory]:
    """Load the history from the bus."""
    history: dict[str, AdvertisementHistory] = {}
    for path, packed_data in managed_objects.items():
        path_str = str(path)
        if not path_str.startswith("/org/bluez/hci"):
            continue

        if not (props := packed_data.get("org.bluez.Device1")):
            continue

        address = props["Address"]
        rssi = props.get("RSSI", MIN_RSSI)

        if (
            prev_history := history.get(address)
        ) and prev_history.advertisement_data.rssi >= rssi:
            continue

        split_path = path_str.split("/")
        adapter = split_path[3]

        if source_adapter and adapter != source_adapter:
            continue

        uuids = props.get("UUIDs", [])
        manufacturer_data = {
            k: bytes(v) for k, v in props.get("ManufacturerData", {}).items()
        }
        device = BLEDevice(
            address,
            props["Alias"],
            {"path": path, "props": props},
        )
        advertisement_data = AdvertisementData(
            local_name=props.get("Name"),
            manufacturer_data=manufacturer_data,
            service_data={k: bytes(v) for k, v in props.get("ServiceData", {}).items()},
            service_uuids=uuids,
            platform_data=props,
            tx_power=props.get("TxPower"),
            rssi=rssi,
        )
        history[device.address] = AdvertisementHistory(
            device, advertisement_data, adapter
        )

    return history



from __future__ import annotations

import logging
from contextlib import ExitStack
from functools import cache
from pathlib import Path
from typing import Any
import asyncio

try:
    from dbus_fast import AuthError, BusType, Message, MessageType, unpack_variants
    from dbus_fast.aio import MessageBus
except (AttributeError, ImportError):
    # dbus_fast is not available on Windows
    AuthError = None  # pragma: no cover
    BusType = None  # pragma: no cover
    Message = None  # pragma: no cover
    MessageType = None  # pragma: no cover
    unpack_variants = None  # pragma: no cover
    MessageBus = None  # pragma: no cover


from .history import AdvertisementHistory, load_history_from_managed_objects
from .util import asyncio_timeout

_LOGGER = logging.getLogger(__name__)

REPLY_TIMEOUT = 8


class BlueZDBusObjects:
    """Fetch and parse BlueZObjects."""

    def __init__(self) -> None:
        """Init the manager."""
        self._packed_managed_objects: dict[str, Any] = {}
        self._unpacked_managed_objects: dict[str, Any] = {}

    async def load(self) -> None:
        """Load from the bus."""
        self._packed_managed_objects = await _get_dbus_managed_objects()
        self._unpacked_managed_objects = {}

    @property
    def adapters(self) -> list[str]:
        """Get adapters."""
        return list(self.adapter_details)

    @property
    def unpacked_managed_objects(self) -> dict[str, Any]:
        """Get unpacked managed objects."""
        if not self._unpacked_managed_objects:
            self._unpacked_managed_objects = unpack_variants(
                self._packed_managed_objects
            )
        return self._unpacked_managed_objects

    @property
    def adapter_details(self) -> dict[str, dict[str, Any]]:
        """Get adapters."""
        return _adapters_from_managed_objects(self.unpacked_managed_objects)

    @property
    def history(self) -> dict[str, AdvertisementHistory]:
        """Get history from managed objects."""
        return load_history_from_managed_objects(self.unpacked_managed_objects)


def _adapters_from_managed_objects(
    managed_objects: dict[str, Any],
) -> dict[str, dict[str, Any]]:
    adapters: dict[str, dict[str, Any]] = {}
    for path, unpacked_data in managed_objects.items():
        path_str = str(path)
        # check that path is exactly /org/bluez/hci<integer>
        if not path_str.startswith("/org/bluez/hci"):
            continue
        if not path_str[14:].isdigit():
            continue
        split_path = path_str.split("/")
        adapter = split_path[3]
        if adapter not in adapters:
            adapters[adapter] = unpacked_data
    return adapters


async def get_bluetooth_adapters() -> list[str]:
    """Return a list of bluetooth adapters."""
    return list(await get_bluetooth_adapter_details())


async def get_bluetooth_adapter_details() -> dict[str, dict[str, Any]]:
    """Return a list of bluetooth adapter with details."""
    results = await _get_dbus_managed_objects()
    return {
        adapter: unpack_variants(packed_data)
        for adapter, packed_data in _adapters_from_managed_objects(results).items()
    }


async def get_dbus_managed_objects() -> dict[str, Any]:
    """Return a list of bluetooth adapter with details."""
    results = await _get_dbus_managed_objects()
    return {path: unpack_variants(packed_data) for path, packed_data in results.items()}


async def _get_dbus_managed_objects() -> dict[str, Any]:
    bus: MessageBus | None = None
    with ExitStack() as stack:
        try:
            bus = MessageBus(bus_type=BusType.SYSTEM)
            stack.callback(bus.disconnect)
            await bus.connect()
        except AuthError as ex:
            _LOGGER.warning(
                "DBus authentication error; make sure the DBus socket "
                "is available and the user has the correct permissions: %s",
                ex,
            )
            return {}
        except FileNotFoundError as ex:
            if is_docker_env():
                _LOGGER.debug(
                    "DBus service not found; docker config may "
                    "be missing `-v /run/dbus:/run/dbus:ro`: %s",
                    ex,
                )
            _LOGGER.debug(
                "DBus service not found; make sure the DBus socket is available: %s",
                ex,
            )
            return {}
        except BrokenPipeError as ex:
            if is_docker_env():
                _LOGGER.debug(
                    "DBus connection broken: %s; try restarting "
                    "`bluetooth`, `dbus`, and finally the docker container",
                    ex,
                )
            _LOGGER.debug(
                "DBus connection broken: %s; try restarting `bluetooth` and `dbus`", ex
            )
            return {}
        except ConnectionRefusedError as ex:
            if is_docker_env():
                _LOGGER.debug(
                    "DBus connection refused: %s; try restarting "
                    "`bluetooth`, `dbus`, and finally the docker container",
                    ex,
                )
            _LOGGER.debug(
                "DBus connection refused: %s; try restarting `bluetooth` and `dbus`", ex
            )
            return {}
        msg = Message(
            destination="org.bluez",
            path="/",
            interface="org.freedesktop.DBus.ObjectManager",
            member="GetManagedObjects",
        )
        try:
            async with asyncio_timeout(REPLY_TIMEOUT):
                reply = await bus.call(msg)
        except EOFError as ex:
            _LOGGER.debug("DBus connection closed: %s", ex)
            return {}
        except asyncio.TimeoutError:
            _LOGGER.debug(
                "Dbus timeout waiting for reply to GetManagedObjects; try restarting "
                "`bluetooth` and `dbus`"
            )
            return {}
        if not reply or reply.message_type != MessageType.METHOD_RETURN:
            _LOGGER.debug(
                "Received an unexpected reply from Dbus while "
                "calling GetManagedObjects on org.bluez: %s",
                reply,
            )
            return {}
        results: dict[str, Any] = reply.body[0]
        return results


@cache
def is_docker_env() -> bool:
    """Return True if we run in a docker env."""
    return Path("/.dockerenv").exists()



import platform

from ..adapters import BluetoothAdapters
from ..const import DEFAULT_ADDRESS, MACOS_DEFAULT_BLUETOOTH_ADAPTER
from ..models import AdapterDetails


class MacOSAdapters(BluetoothAdapters):
    """Class for getting the bluetooth adapters on a MacOS system."""

    @property
    def adapters(self) -> dict[str, AdapterDetails]:
        """Get the adapter details."""
        return {
            MACOS_DEFAULT_BLUETOOTH_ADAPTER: AdapterDetails(
                address=DEFAULT_ADDRESS,
                sw_version=platform.release(),
                passive_scan=False,
                manufacturer="Apple",
                product="Unknown MacOS Model",
                vendor_id="Unknown",
                product_id="Unknown",
                adapter_type=None,  # Unknown for macOS
                advertise=True,
                powered=True,
            )
        }

    @property
    def default_adapter(self) -> str:
        """Get the default adapter."""
        return MACOS_DEFAULT_BLUETOOTH_ADAPTER



"""Utils."""

from __future__ import annotations

import sys

from .const import DEFAULT_ADDRESS
from .models import (
    ADAPTER_PRODUCT,
    ADAPTER_PRODUCT_ID,
    ADAPTER_VENDOR_ID,
    AdapterDetails,
)

if sys.version_info[:2] < (3, 11):
    from async_timeout import timeout as asyncio_timeout  # noqa: F401
else:
    from asyncio import timeout as asyncio_timeout  # noqa: F401


def adapter_human_name(adapter: str, address: str) -> str:
    """Return a human readable name for the adapter."""
    return adapter if address == DEFAULT_ADDRESS else f"{adapter} ({address})"


def adapter_unique_name(adapter: str, address: str) -> str:
    """Return a unique name for the adapter."""
    return adapter if address == DEFAULT_ADDRESS else address


def adapter_model(details: AdapterDetails) -> str:
    """Return a model for the adapter."""
    if (vendor_id := details.get(ADAPTER_VENDOR_ID)) and vendor_id != "Unknown":
        return f"{details[ADAPTER_PRODUCT]} ({vendor_id}:{details[ADAPTER_PRODUCT_ID]})"
    return details.get(ADAPTER_PRODUCT) or "Unknown"



import asyncio

from bluetooth_adapters import get_adapters


async def go() -> None:
    bluetooth_adapters = get_adapters()
    await bluetooth_adapters.refresh()
    print(bluetooth_adapters.adapters)


asyncio.run(go())



__version__ = "2.4.0"


from platform import system

from .adapters import BluetoothAdapters
from .const import (
    DEFAULT_ADDRESS,
    DEFAULT_CONNECTION_SLOTS,
    MACOS_DEFAULT_BLUETOOTH_ADAPTER,
    UNIX_DEFAULT_BLUETOOTH_ADAPTER,
    WINDOWS_DEFAULT_BLUETOOTH_ADAPTER,
)

if system() != "Windows":
    from .dbus import (
        BlueZDBusObjects,
        get_bluetooth_adapter_details,
        get_bluetooth_adapters,
        get_dbus_managed_objects,
    )

from .history import AdvertisementHistory, load_history_from_managed_objects
from .mac_lookup import get_manufacturer_from_mac
from .models import (
    ADAPTER_ADDRESS,
    ADAPTER_ADVERTISE,
    ADAPTER_CONNECTION_SLOTS,
    ADAPTER_HW_VERSION,
    ADAPTER_MANUFACTURER,
    ADAPTER_PASSIVE_SCAN,
    ADAPTER_POWERED,
    ADAPTER_PRODUCT,
    ADAPTER_PRODUCT_ID,
    ADAPTER_SW_VERSION,
    ADAPTER_TYPE,
    ADAPTER_VENDOR_ID,
    AdapterDetails,
)
from .storage import (
    DiscoveredDeviceAdvertisementData,
    DiscoveredDeviceAdvertisementDataDict,
    DiscoveryStorageType,
    discovered_device_advertisement_data_from_dict,
    discovered_device_advertisement_data_to_dict,
    expire_stale_scanner_discovered_device_advertisement_data,
)
from .systems import get_adapters
from .systems.linux_hci import get_adapters_from_hci
from .util import adapter_human_name, adapter_model, adapter_unique_name

__all__ = [
    "ADAPTER_ADDRESS",
    "ADAPTER_ADVERTISE",
    "ADAPTER_CONNECTION_SLOTS",
    "ADAPTER_HW_VERSION",
    "ADAPTER_MANUFACTURER",
    "ADAPTER_PASSIVE_SCAN",
    "ADAPTER_POWERED",
    "ADAPTER_PRODUCT",
    "ADAPTER_PRODUCT_ID",
    "ADAPTER_SW_VERSION",
    "ADAPTER_TYPE",
    "ADAPTER_VENDOR_ID",
    "DEFAULT_ADDRESS",
    "DEFAULT_CONNECTION_SLOTS",
    "MACOS_DEFAULT_BLUETOOTH_ADAPTER",
    "UNIX_DEFAULT_BLUETOOTH_ADAPTER",
    "WINDOWS_DEFAULT_BLUETOOTH_ADAPTER",
    "AdapterDetails",
    "AdvertisementHistory",
    "BlueZDBusObjects",
    "BluetoothAdapters",
    "DiscoveredDeviceAdvertisementData",
    "DiscoveredDeviceAdvertisementDataDict",
    "DiscoveryStorageType",
    "adapter_human_name",
    "adapter_model",
    "adapter_unique_name",
    "discovered_device_advertisement_data_from_dict",
    "discovered_device_advertisement_data_to_dict",
    "expire_stale_scanner_discovered_device_advertisement_data",
    "get_adapters",
    "get_adapters_from_hci",
    "get_bluetooth_adapter_details",
    "get_bluetooth_adapters",
    "get_dbus_managed_objects",
    "get_manufacturer_from_mac",
    "load_history_from_managed_objects",
]



"""Tests for the raw kernel HCI device query in systems/linux_hci.py.

These exercise ``get_adapters_from_hci`` directly by faking the ``socket`` and
``fcntl`` module references it holds, so the ctypes struct parsing and ioctl
plumbing are covered without any real Bluetooth hardware. This is the code path
that issue #118 exposed as both fragile and untested.
"""

from __future__ import annotations

import ctypes
from collections.abc import Callable
from typing import Any
from unittest.mock import MagicMock, patch

import pytest

from bluetooth_adapters.systems import linux_hci
from bluetooth_adapters.systems.linux_hci import (
    HCIGETDEVINFO,
    HCIGETDEVLIST,
    get_adapters_from_hci,
)


def _make_ioctl(
    devices: dict[int, tuple[str, list[int]]]
    | dict[int, tuple[str, list[int], int, list[int]]],
) -> Callable[[int, int, Any], int]:
    """Build a fake ``fcntl.ioctl`` that populates the ctypes buffers.

    ``devices`` maps a device id to ``(name, bdaddr_bytes)`` or
    ``(name, bdaddr_bytes, flags, features)`` where ``bdaddr_bytes`` is the
    6-byte little-endian address as stored by the kernel (``bdaddr_t.__str__``
    reverses it when rendering) and ``features`` is the 8-byte LMP feature
    mask. When ``flags``/``features`` are omitted they default to zero.
    """

    def ioctl(_fd: int, request: int, arg: Any) -> int:
        if request == HCIGETDEVLIST:
            arg.dev_num = len(devices)
            for i, dev_id in enumerate(devices):
                arg.dev_req[i].dev_id = dev_id
        elif request == HCIGETDEVINFO:
            entry = devices[arg.dev_id]
            name, bdaddr = entry[0], entry[1]
            arg.name = name.encode()
            arg.bdaddr.b = (ctypes.c_uint8 * 6)(*bdaddr)
            if len(entry) == 4:
                arg.flags = entry[2]
                arg.features = (ctypes.c_uint8 * 8)(*entry[3])
        return 0

    return ioctl


def test_get_adapters_from_hci_parses_devices() -> None:
    """Two adapters are returned with decoded names and addresses."""
    devices = {
        0: ("hci0", [0x04, 0x71, 0xDA, 0x7D, 0x1A, 0x00]),
        1: ("hci1", [0x05, 0x71, 0xDA, 0x7D, 0x1A, 0x00]),
    }
    fake_fcntl = MagicMock()
    fake_fcntl.ioctl.side_effect = _make_ioctl(devices)

    with (
        patch.object(linux_hci, "fcntl", fake_fcntl),
        patch.object(linux_hci, "socket"),
    ):
        out = get_adapters_from_hci()

    assert set(out) == {0, 1}
    assert out[0]["name"] == "hci0"
    assert out[0]["bdaddr"] == "00:1A:7D:DA:71:04"
    assert out[1]["name"] == "hci1"
    assert out[1]["bdaddr"] == "00:1A:7D:DA:71:05"


def test_get_adapters_from_hci_parses_powered_and_advertise() -> None:
    """Powered (HCI_UP) and advertise (LMP_LE) bits are decoded correctly."""
    devices = {
        # Powered, LE capable: HCI_UP bit 0 set, features[4] has LMP_LE (0x40).
        0: (
            "hci0",
            [0x04, 0x71, 0xDA, 0x7D, 0x1A, 0x00],
            0x01,
            [0x00, 0x00, 0x00, 0x00, 0x40, 0x00, 0x00, 0x00],
        ),
        # Down, no LE: flags and features all zero.
        1: (
            "hci1",
            [0x05, 0x71, 0xDA, 0x7D, 0x1A, 0x00],
            0x00,
            [0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00],
        ),
        # Powered but no LE; also exercises other flag bits being ignored.
        2: (
            "hci2",
            [0x06, 0x71, 0xDA, 0x7D, 0x1A, 0x00],
            0xFF,
            [0xFF, 0xFF, 0xFF, 0xFF, 0xBF, 0xFF, 0xFF, 0xFF],
        ),
    }
    fake_fcntl = MagicMock()
    fake_fcntl.ioctl.side_effect = _make_ioctl(devices)

    with (
        patch.object(linux_hci, "fcntl", fake_fcntl),
        patch.object(linux_hci, "socket"),
    ):
        out = get_adapters_from_hci()

    assert out[0]["powered"] is True
    assert out[0]["advertise"] is True
    assert out[1]["powered"] is False
    assert out[1]["advertise"] is False
    assert out[2]["powered"] is True
    assert out[2]["advertise"] is False


def test_get_adapters_from_hci_empty_when_no_devices() -> None:
    """No HCI devices yields an empty mapping (and only one ioctl call)."""
    fake_fcntl = MagicMock()
    fake_fcntl.ioctl.side_effect = _make_ioctl({})

    with (
        patch.object(linux_hci, "fcntl", fake_fcntl),
        patch.object(linux_hci, "socket"),
    ):
        out = get_adapters_from_hci()

    assert out == {}
    # Only the device-list ioctl runs; no per-device queries.
    assert fake_fcntl.ioctl.call_count == 1


def test_get_adapters_from_hci_closes_socket() -> None:
    """The socket is always closed, even on the happy path."""
    fake_fcntl = MagicMock()
    fake_fcntl.ioctl.side_effect = _make_ioctl({})
    fake_socket_mod = MagicMock()
    sock = fake_socket_mod.socket.return_value

    with (
        patch.object(linux_hci, "fcntl", fake_fcntl),
        patch.object(linux_hci, "socket", fake_socket_mod),
    ):
        get_adapters_from_hci()

    sock.close.assert_called_once()


def test_get_adapters_from_hci_raises_without_fcntl() -> None:
    """On platforms without fcntl (Windows) a RuntimeError is raised."""
    with patch.object(linux_hci, "fcntl", None):
        with pytest.raises(RuntimeError, match="fcntl is not available"):
            get_adapters_from_hci()


def test_get_adapters_from_hci_oserror_returns_empty() -> None:
    """An OSError from the kernel ioctl is swallowed and returns {}."""
    fake_fcntl = MagicMock()
    fake_fcntl.ioctl.side_effect = OSError("HCIGETDEVLIST failed")

    with (
        patch.object(linux_hci, "fcntl", fake_fcntl),
        patch.object(linux_hci, "socket"),
    ):
        out = get_adapters_from_hci()

    assert out == {}


def test_get_adapters_from_hci_negative_ret_returns_empty() -> None:
    """A negative ioctl return (raised as OSError internally) yields {}."""
    fake_fcntl = MagicMock()
    fake_fcntl.ioctl.return_value = -1

    with (
        patch.object(linux_hci, "fcntl", fake_fcntl),
        patch.object(linux_hci, "socket"),
    ):
        out = get_adapters_from_hci()

    assert out == {}


def test_get_adapters_from_hci_unexpected_error_returns_empty() -> None:
    """A non-OSError exception is logged and returns {} rather than raising."""
    fake_fcntl = MagicMock()
    fake_fcntl.ioctl.side_effect = ValueError("boom")

    with (
        patch.object(linux_hci, "fcntl", fake_fcntl),
        patch.object(linux_hci, "socket"),
    ):
        out = get_adapters_from_hci()

    assert out == {}


def test_bdaddr_t_str_reverses_bytes() -> None:
    """bdaddr_t renders the kernel's little-endian bytes as big-endian text."""
    addr = linux_hci.bdaddr_t()
    addr.b = (ctypes.c_uint8 * 6)(0x04, 0x71, 0xDA, 0x7D, 0x1A, 0x00)
    assert str(addr) == "00:1A:7D:DA:71:04"



"""Base class for Bluetooth adapters."""

from __future__ import annotations

from abc import ABC, abstractmethod

from .history import AdvertisementHistory
from .models import AdapterDetails


class BluetoothAdapters(ABC):
    """Class for getting the bluetooth adapters on a system."""

    async def refresh(self) -> None:
        """Refresh the adapters."""

    @property
    def history(self) -> dict[str, AdvertisementHistory]:
        """Get the history."""
        return {}

    @property
    @abstractmethod
    def adapters(self) -> dict[str, AdapterDetails]:
        """Get the adapter details."""

    @property
    @abstractmethod
    def default_adapter(self) -> str:
        """Get the default adapter."""



"""Constants for bluetooth adapters."""

from typing import Final

WINDOWS_DEFAULT_BLUETOOTH_ADAPTER: Final = "bluetooth"
MACOS_DEFAULT_BLUETOOTH_ADAPTER: Final = "Core Bluetooth"
UNIX_DEFAULT_BLUETOOTH_ADAPTER: Final = "hci0"

# Some operating systems hide the adapter address for privacy reasons (ex MacOS)
DEFAULT_ADDRESS: Final = "00:00:00:00:00:00"

EMPTY_MAC_ADDRESS: Final = "00:00:00:00:00:00"

DEFAULT_CONNECTION_SLOTS: Final = 5
